from tcr_embedding.models.architectures.mlp import MLP


def build_mlp_encoder(params, xdim, hdim):
    return MLP(xdim, hdim, params['gene_hidden'], params['activation'], params['activation'], params['dropout'],
               params['batch_norm'], regularize_last_layer=True)


def build_mlp_decoder(params, xdim, hdim):
    return MLP(hdim, xdim, params['gene_hidden'][::-1], params['activation'], 'linear',
               params['dropout'], params['batch_norm'], regularize_last_layer=False)



import torch
import torch.nn as nn


class BiGRUEncoder(nn.Module):
	def __init__(self, params, hdim, num_seq_labels):
		"""

		:param params: hyperparameters as dict
		:param hdim: output feature dimension
		:param num_seq_labels: number of aa labels, input dim
		"""
		super(BiGRUEncoder, self).__init__()
		self.embedding_dim = params['embedding_dim']
		self.hidden_size = params['hidden_size']
		self.num_layers = params['num_layers']
		self.dropout = params['dropout']
		self.bidirectional = params['bidirectional']
		self.num_seq_labels = num_seq_labels
		self.hdim = hdim

		self.embedding = nn.Embedding(num_embeddings=num_seq_labels, embedding_dim=params['embedding_dim'], padding_idx=0)

		self.gru = nn.GRU(
			input_size=params['embedding_dim'],
			hidden_size=params['hidden_size'],
			num_layers=params['num_layers'],
			batch_first=True,
			dropout=params['dropout'],
			bidirectional=params['bidirectional']
		)

		input_size = params['hidden_size'] * params['num_layers'] * 2 if params['bidirectional'] else params['hidden_size'] * params['num_layers']

		self.output_layer = nn.Linear(input_size, hdim)

	def forward(self, tcr_seq, tcr_len):
		"""
		:param tcr_seq: torch.LongTensor, shape=[batch_size, max_seq_len]
		:param tcr_len: torch.LongTensor, shape=[batch_size, 1] holds the real sequence length (without padding)
		:return: torch.Tensor, shape=[batch_size, hdim]
		"""

		# initialize hidden state
		num_layers = self.num_layers * 2 if self.bidirectional else self.num_layers
		batch_size = tcr_seq.shape[0]
		h0 = torch.randn(num_layers, batch_size, self.hidden_size).to(tcr_seq.device)

		x = self.embedding(tcr_seq)  # [batch_size, max_seq_len, embedding_dim]
		x = torch.nn.utils.rnn.pack_padded_sequence(x, tcr_len, batch_first=True, enforce_sorted=False)
		tokens, hidden = self.gru(x, h0)  # hidden.shape=[num_layer(*2 if bidirectional), batch_size, hidden_size]
		# tokens, _ = torch.nn.utils.rnn.pad_packed_sequence(tokens, batch_first=True)  # output of each token not needed

		hidden = hidden.permute(1, 0, 2)  # shape=[batch_size, num_layers(*2 if biGRU), hidden_size]
		hidden = hidden.flatten(1)  # shape=[batch*num_layers, hidden_size]

		output = self.output_layer(hidden)

		return output


class BiGRUDecoder(nn.Module):
	def __init__(self, params, input_dim, num_seq_labels):
		"""

		:param params: hyperparameters as dict
		:param hdim: input feature dimension
		:param num_seq_labels: number of aa labels, output dim
		"""
		super(BiGRUDecoder, self).__init__()
		self.hidden_size = params['hidden_size']
		self.num_layers = params['num_layers']
		self.dropout = params['dropout']
		self.bidirectional = params['bidirectional']
		self.num_seq_labels = num_seq_labels
		self.input_dim = input_dim
		self.teacher_forcing = params['teacher_forcing']

		self.embedding = nn.Embedding(num_embeddings=num_seq_labels, embedding_dim=params['embedding_dim'], padding_idx=0)
		self.hidden_state_layer = nn.Linear(input_dim, params['hidden_size'] * params['num_layers'])

		self.gru = nn.GRU(
			input_size=params['embedding_dim'],
			hidden_size=params['hidden_size'],
			num_layers=params['num_layers'],
			batch_first=True,
			dropout=params['dropout'],
			bidirectional=False
		)

		self.output_layer = nn.Linear(params['hidden_size'], num_seq_labels)
		# self.softmax = nn.Softmax(dim=-1)  # not needed using nn.CrossEntropyLoss


	def forward(self, shared_hidden_state, gt_tcr_seq):
		"""
		Forward pass including autoregressively predicting the sequence using a for loop
		:param shared_hidden_state: shared hidden state after shared_decoder, torch.FloatTensor with shape [batch_size, input_dim]
		:param gt_tcr_seq: torch.LongTensor, shape=[batch_size, max_seq_len] ground truth tcr-seq, same as encoder input, only needed for shape and start symbol
		:return:
		"""
		decoder_raw_output = []
		batch_size, seq_len = gt_tcr_seq.shape

		hidden_state = self.hidden_state_layer(shared_hidden_state)  # shape=[batch_size, num_layers*hidden_size]
		hidden_state = hidden_state.reshape(batch_size, self.num_layers, self.hidden_size)  # shape=[batch_size, num_layers, hidden_size]
		hidden_state = hidden_state.permute(1, 0, 2)  # switch batch and layer dim, shape=[num_layers, batch_size, hidden_size]
		hidden_state = hidden_state.contiguous()  # permute returns a non-contiguous Tensor

		use_teacher_forcing = True if torch.rand(1) < self.teacher_forcing else False
		if use_teacher_forcing:
			input = gt_tcr_seq[:, :-1]  # all timesteps except last one
			embedding = self.embedding(input)

			output, hidden_state = self.gru(embedding, hidden_state)
			decoder_raw_output = self.output_layer(output)

		else:
			# Initialize first input and hidden state
			input = gt_tcr_seq[:, 0].unsqueeze(1)  # start symbol
			for step in range(seq_len-1):
				output, hidden_state = self.step(input, hidden_state)
				input = torch.argmax(output, dim=-1)  # only needs to be seq_len=1 as hidden_state contains the previous information

				decoder_raw_output.append(output)
			decoder_raw_output = torch.cat(decoder_raw_output, dim=1)

		return decoder_raw_output

	def step(self, input, hidden_state):
		"""
		A single forward step to predict the next token
		:param input: torch.Tensor with shape [batch_size, seq_len]
		:param hidden_state: torch.Tensor with shape [num_layers, batch_size, hidden_size]
		:return: output, hidden_state
		"""
		embedding = self.embedding(input)
		output, hidden_state = self.gru(embedding, hidden_state)
		output = self.output_layer(output)
		# output = self.softmax(output)  # not needed using nn.CrossEntropyLoss

		return output, hidden_state

	def predict(self, shared_hidden_state, gt_tcr_seq):
		"""
		Only feed in own prediction, don't use teacher forcing here
		:param shared_hidden_state:
		:param gt_tcr_seq:
		:return:
		"""
		decoder_raw_output = []
		batch_size, seq_len = gt_tcr_seq.shape

		hidden_state = self.hidden_state_layer(shared_hidden_state)  # shape=[batch_size, num_layers*hidden_size]
		hidden_state = hidden_state.reshape(batch_size, self.num_layers, self.hidden_size)  # shape=[batch_size, num_layers, hidden_size]
		hidden_state = hidden_state.permute(1, 0, 2)  # switch batch and layer dim, shape=[num_layers, batch_size, hidden_size]
		hidden_state = hidden_state.contiguous()  # permute returns a non-contiguous Tensor

		# Initialize first input and hidden state
		input = gt_tcr_seq[:, 0].unsqueeze(1)  # start symbol
		for step in range(seq_len-1):
			output, hidden_state = self.step(input, hidden_state)
			input = torch.argmax(output, dim=-1)  # only needs to be seq_len=1 as hidden_state contains the previous information

			decoder_raw_output.append(output)
		decoder_raw_output = torch.cat(decoder_raw_output, dim=1)

		return decoder_raw_output



# A Variational Information Bottleneck Approach to Multi-Omics Data Integration
import torch
import torch.nn as nn
import scanpy as sc

from tcr_embedding.models.architectures.transformer import TransformerEncoder, TransformerDecoder
from tcr_embedding.models.architectures.mlp import MLP
from tcr_embedding.models.architectures.mlp_scRNA import build_mlp_encoder, build_mlp_decoder
from tcr_embedding.models.vae_base_model import VAEBaseModel
from tcr_embedding.dataloader.DataLoader import initialize_prediction_loader


class MoEModelTorch(nn.Module):
	def __init__(self, tcr_params, rna_params, joint_params):
		super(MoEModelTorch, self).__init__()
		self.beta_only = 'beta_only' in tcr_params and tcr_params['beta_only']
		self.amount_chains = 1 if self.beta_only else 2

		xdim = rna_params['xdim']
		hdim = joint_params['hdim']
		num_conditional_labels = joint_params['num_conditional_labels']
		cond_dim = joint_params['cond_dim']
		cond_input = joint_params['cond_input']
		zdim = joint_params['zdim']
		shared_hidden = joint_params['shared_hidden']
		activation = joint_params['activation']
		dropout = joint_params['dropout']
		batch_norm = joint_params['batch_norm']
		use_embedding_for_cond = joint_params['use_embedding_for_cond'] if 'use_embedding_for_cond' in joint_params else True

		num_seq_labels = tcr_params['num_seq_labels']

		if not self.beta_only:
			self.alpha_encoder = TransformerEncoder(tcr_params, hdim // 2, num_seq_labels)
			self.alpha_decoder = TransformerDecoder(tcr_params, hdim, num_seq_labels)

		self.beta_encoder = TransformerEncoder(tcr_params, hdim // self.amount_chains, num_seq_labels)
		self.beta_decoder = TransformerDecoder(tcr_params, hdim, num_seq_labels)

		self.rna_encoder = build_mlp_encoder(rna_params, xdim, hdim)
		self.rna_decoder = build_mlp_decoder(rna_params, xdim, hdim)

		if cond_dim > 0:
			if use_embedding_for_cond:
				self.cond_emb = torch.nn.Embedding(num_conditional_labels, cond_dim)
			else:  # use one hot encoding
				self.cond_emb = None
				cond_dim = num_conditional_labels
		self.cond_input = cond_input
		self.use_embedding_for_cond = use_embedding_for_cond
		self.num_conditional_labels = num_conditional_labels
		cond_input_dim = cond_dim if cond_input else 0

		self.tcr_vae_encoder = MLP(hdim + cond_input_dim, zdim * 2, shared_hidden, activation, 'linear', dropout,
								   batch_norm, regularize_last_layer=False)
		self.tcr_vae_decoder = MLP(zdim + cond_dim, hdim, shared_hidden[::-1], activation, activation, dropout,
								   batch_norm, regularize_last_layer=True)

		self.rna_vae_encoder = MLP(hdim + cond_input_dim, zdim * 2, shared_hidden, activation, 'linear', dropout,
								   batch_norm, regularize_last_layer=False)
		self.rna_vae_decoder = MLP(zdim + cond_dim, hdim, shared_hidden[::-1], activation, activation, dropout,
								   batch_norm, regularize_last_layer=True)

		# used for NB loss
		self.theta = torch.nn.Parameter(torch.randn(xdim))

	def forward(self, rna, tcr, tcr_len, conditional=None):
		"""
		Forward pass of autoencoder
		:param rna: torch.Tensor shape=[batch_size, num_genes]
		:param tcr: torch.Tensor shape=[batch_size, seq_len, feature_dim]
		:param tcr_len: torch.Tensor shape=[batch_size]
		:param conditional: torch.Tensor shape=[batch_size, n_cond] one-hot-encoded conditional covariates
		:return:
			z: list of sampled latent variable zs. z = [z_rna, z_tcr, z_joint]
			mu: list of predicted means mu. mu = [mu_rna, mu_tcr, mu_joint]
			logvar: list of predicted logvars. logvar = [logvar_rna, logvar_tcr, logvar_joint]
			rna_pred: list of reconstructed rna. rna_pred = [rna_pred using z_rna, rna_pred using z_joint]
			tcr_pred: list of reconstructed tcr. tcr_pred = [tcr_pred using z_tcr, tcr_pred using z_joint]
		"""
		if conditional is not None:
			cond_emb_vec = self.cond_emb(conditional) if self.use_embedding_for_cond else torch.nn.functional.one_hot(conditional, self.num_conditional_labels)
		# Encode TCR

		beta_seq = tcr[:, tcr.shape[1] // 2 * (self.amount_chains == 2):]
		beta_len = tcr_len[:, self.amount_chains - 1]
		h_beta = self.beta_encoder(beta_seq, beta_len)  # shape=[batch_size, hdim//2]

		if not self.beta_only:
			alpha_seq = tcr[:, :tcr.shape[1] // 2]
			alpha_len = tcr_len[:, 0]
			h_alpha = self.alpha_encoder(alpha_seq, alpha_len)  # shape=[batch_size, hdim//2]
			h_tcr = torch.cat([h_alpha, h_beta], dim=-1)  # shape=[batch_size, hdim]
		else:
			h_tcr = torch.cat([h_beta], dim=-1)

		if conditional is not None and self.cond_input:
			h_tcr = torch.cat([h_tcr, cond_emb_vec], dim=1)  # shape=[batch_size, hdim+n_cond]

		# Encode RNA
		h_rna = self.rna_encoder(rna)  # shape=[batch_size, hdim]
		if conditional is not None and self.cond_input:
			h_rna = torch.cat([h_rna, cond_emb_vec], dim=1)  # shape=[batch_size, hdim+n_cond]

		# Predict Latent space
		z_rna_ = self.rna_vae_encoder(h_rna)  # shape=[batch_size, zdim*2]
		mu_rna, logvar_rna = z_rna_[:, :z_rna_.shape[1] // 2], z_rna_[:, z_rna_.shape[1] // 2:]
		z_rna = self.reparameterize(mu_rna, logvar_rna)  # shape=[batch_size, zdim]

		z_tcr_ = self.tcr_vae_encoder(h_tcr)  # shape=[batch_size, zdim*2]
		mu_tcr, logvar_tcr = z_tcr_[:, :z_tcr_.shape[1] // 2], z_tcr_[:, z_tcr_.shape[1] // 2:]
		z_tcr = self.reparameterize(mu_tcr, logvar_tcr)  # shape=[batch_size, zdim]

		z = [z_rna, z_tcr]
		mu = [mu_rna, mu_tcr]
		logvar = [logvar_rna, logvar_tcr]

		# Reconstruction
		rna_pred = []
		for z_ in z:
			if conditional is not None:
				z_ = torch.cat([z_, cond_emb_vec], dim=1)  # shape=[batch_size, hdim+n_cond]
			f_rna = self.rna_vae_decoder(z_)
			rna_pred.append(self.rna_decoder(f_rna))
		tcr_pred = []
		for z_ in z:
			if conditional is not None:
				z_ = torch.cat([z_, cond_emb_vec], dim=1)  # shape=[batch_size, hdim+n_cond]
			f_tcr = self.tcr_vae_decoder(z_)
			beta_pred = self.beta_decoder(f_tcr, beta_seq)

			if not self.beta_only:
				alpha_pred = self.alpha_decoder(f_tcr, alpha_seq)
				tcr_pred.append(torch.cat([alpha_pred, beta_pred], dim=1))
			else:
				tcr_pred.append(torch.cat([beta_pred], dim=1))
		return z, mu, logvar, rna_pred, tcr_pred

	def reparameterize(self, mu, log_var):
		"""
		https://debuggercafe.com/getting-started-with-variational-autoencoder-using-pytorch/
		:param mu: mean from the encoder's latent space
		:param log_var: log variance from the encoder's latent space
		"""
		std = torch.exp(0.5 * log_var)  # standard deviation
		eps = torch.randn_like(std)  # `randn_like` as we need the same size
		z = mu + (eps * std)  # sampling as if coming from the input space
		return z

	def predict_transcriptome(self, z_shared, conditional=None):
		"""
		Predict the transcriptome connected to an shared latent space
		:param z_shared: torch.tensor, shared latent representation
		:param conditional:
		:return: torch.tensor, transcriptome profile
		"""
		if conditional is not None:  # more efficient than doing two concatenations
			cond_emb_vec = self.cond_emb(conditional)
			z_shared = torch.cat([z_shared, cond_emb_vec], dim=-1)  # shape=[batch_size, zdim+cond_dim]
		transcriptome_pred = self.rna_vae_decoder(z_shared)
		transcriptome_pred = self.rna_decoder(transcriptome_pred)
		return transcriptome_pred

	def get_latent_from_z(self, z):
		z = 0.5 * (z[0] + z[1])
		return z


class MoEModel(VAEBaseModel):
	def __init__(self,
				 adata,
				 params_architecture,
				 balanced_sampling='clonotype',
				 metadata=None,
				 conditional=None,
				 optimization_mode_params=None,
				 label_key=None,
				 device=None
				 ):
		super(MoEModel, self).__init__(adata, params_architecture, balanced_sampling, metadata,
									   conditional, optimization_mode_params, label_key, device)
		self.model_type = 'moe'

		self.params_tcr['max_tcr_length'] = adata.obsm['alpha_seq'].shape[1]
		self.params_tcr['num_seq_labels'] = len(self.aa_to_id)

		self.params_rna['xdim'] = adata[0].X.shape[1]

		num_conditional_labels = 0
		cond_dim = 0
		if self.conditional is not None:
			if self.conditional in adata.obsm:
				num_conditional_labels = adata.obsm[self.conditional].shape[1]
			else:
				num_conditional_labels = len(adata.obs[self.conditional].unique())
			if 'c_embedding_dim' not in self.params_joint:
				cond_dim = 20
			else:
				cond_dim = self.params_joint['c_embedding_dim']
		self.params_joint['num_conditional_labels'] = num_conditional_labels
		self.params_joint['cond_dim'] = cond_dim
		self.params_joint['cond_input'] = conditional is not None

		self.model = MoEModelTorch(self.params_tcr, self.params_rna, self.params_joint)

	def calculate_loss(self, rna_pred, rna, tcr_pred, tcr):
		rna_loss = self.loss_function_rna(rna_pred[0], rna) + self.loss_function_rna(rna_pred[1], rna)
		rna_loss *= 0.5 * self.loss_weights[0]

		# For GRU and Transformer, as they don't predict start token for alpha and beta chain, so amount of chains used
		if tcr_pred[0].shape[1] == tcr.shape[1] - self.model.amount_chains:
			mask = torch.ones_like(tcr).bool()
			mask[:, [0]] = False
			if not self.beta_only:
				mask[:, [mask.shape[1] // 2]] = False
			tcr_loss = self.loss_function_tcr(tcr_pred[0].flatten(end_dim=1), tcr[mask].flatten())
			if not 'beta_only' in self.params_tcr or self.params_tcr['beta_only']:
				tcr_loss += self.loss_function_tcr(tcr_pred[1].flatten(end_dim=1), tcr[mask].flatten())
			tcr_loss *= 0.5 * self.loss_weights[1]
		else:  # For CNN, as it predicts start token
			tcr_loss = (self.loss_function_tcr(tcr_pred[0].flatten(end_dim=1), tcr.flatten()) +
						self.loss_function_tcr(tcr_pred[1].flatten(end_dim=1), tcr.flatten()))
			tcr_loss *= 0.5 * self.loss_weights[1]
		return rna_loss, tcr_loss

	def calculate_kld_loss(self, mu, logvar, epoch):
		kld_loss = (self.loss_function_kld(mu[0], logvar[0]) + self.loss_function_kld(mu[1], logvar[1]))
		kld_loss *= 0.5 * self.loss_weights[2] * self.get_kl_annealing_factor(epoch)
		z = 0.5 * (mu[0] + mu[1])
		return kld_loss, z

	def get_latent_unimodal(self, adata, metadata, modality, return_mean=True):
		"""
		Get unimodal latent either from RNA or TCR only
		:param adata:
		:param metadata: list of str, list of metadata that is needed, not really useful at the moment
		:param return_mean: bool, calculate latent space without sampling
		:return: adata containing embedding vector in adata.X for each cell and the specified metadata in adata.obs
		"""
		data_embed = initialize_prediction_loader(adata, metadata, self.batch_size, beta_only=self.beta_only)

		zs = []
		with torch.no_grad():
			self.model = self.model.to(self.device)
			self.model.eval()
			for rna, tcr, seq_len, _, labels, conditional in data_embed:
				rna = rna.to(self.device)
				tcr = tcr.to(self.device)

				if self.conditional is not None:
					conditional = conditional.to(self.device)
				else:
					conditional = None
				z, mu, _, _, _ = self.model(rna, tcr, seq_len, conditional)
				if return_mean:
					z = mu
				if modality == 'RNA':
					z = z[0]
				else:
					z = z[1]
				z = sc.AnnData(z.detach().cpu().numpy())
				# z.obs[metadata] = np.array(metadata_batch).T
				zs.append(z)
		latent = sc.AnnData.concatenate(*zs)
		latent.obs.index = adata.obs.index
		latent.obs[metadata] = adata.obs[metadata]
		return latent



import operator
from tcr_embedding.evaluation.WrapperFunctions import get_model_prediction_function
from tcr_embedding.evaluation.kNN import run_knn_within_set_evaluation


def report_pseudo_metric(adata, model, optimization_mode_params, epoch, comet):
    """
    Calculate a pseudo metric based on kNN of multiple meta information
    :param epoch:
    :return:
    """
    prediction_label = optimization_mode_params['prediction_labels']

    labels = prediction_label.copy()
    if isinstance(prediction_label, dict):
        labels = list(prediction_label.keys())

    test_embedding_func = get_model_prediction_function(model, do_adata=True, metadata=labels)
    summary = run_knn_within_set_evaluation(adata, test_embedding_func, labels, subset='val')

    if isinstance(prediction_label, list):
        summary['pseudo_metric'] = sum(summary.values())
    else:
        summary['pseudo_metric'] = sum([summary[f'weighted_f1_{label}'] * prediction_label[label] for label in prediction_label])

    if comet is not None:
        comet.log_metrics(summary, epoch=epoch)

    return summary['pseudo_metric'], operator.gt



import numpy as np


def get_model_prediction_function(model, do_adata=False, metadata=None):
    """
    Wrapper function for our embedding models
    :param model: trained pytorch model
    :param do_adata: return the adata object (else numpy array)
    :param metadata: add these .obs from the original to the created anndata object
    :return: function for calculating the latent space of cells in an anndata object
    """
    def prediction_function(data):
        """
        Calculates the latent space for a dataset
        :param data: anndata object containing the cell data
        :return: numpy array (num_cells, hidden_dim) latent embedding for each cell
        """
        metadata_tmp = metadata if metadata is not None else []
        latent_space = model.get_latent(data, metadata=metadata_tmp, return_mean=True)
        if do_adata:
            return latent_space
        latent_space = latent_space.X
        return latent_space
    return prediction_function


def get_random_prediction_function(hidden_dim):
    """
    Wrapper for calculating a random latent space without input output dependencies as worst case baseline
    :param hidden_dim: dimensionality of the latent space
    :return: function which calculates the latent space
    """
    def prediction_function(data):
        """
        Calculates a random latent space independent of input data
        :param data: anndata object containing the cell data
        :return: numpy array (num_cells, hidden_dim) representing the latent space
        """
        n = len(data.obs)
        np.random.seed(29031995)
        latent_embedding = np.zeros(shape=(n, hidden_dim))
        for i in range(n):
            latent_embedding[i, :] = np.random.random(hidden_dim)
        return latent_embedding
    return prediction_function



import tcr_embedding.evaluation.Metrics as Metrics
import scanpy as sc


def run_knn_within_set_evaluation(data_full, embedding_function, prediction_labels, subset='val'):
    """
    Function for evaluating the embedding quality based upon kNN (k=1) prediction inside the set
    :param data_full: anndata object containing the full cell data (TCR + Genes) (train, val, test)
    :param embedding_function: function calculating the latent space for a single input
    :param prediction_labels: str or list[str], column name in data_full.obs that we want to predict
    :param subset: str to choose the set for testing
    :return: dictionary {metric: summary} containing the evaluation scores
    """
    if type(subset) == str:
        subset = [subset]
    if type(prediction_labels) == str:
        prediction_labels = [prediction_labels]
    data_test = data_full[data_full.obs['set'].isin(subset)]
    latent_test = embedding_function(data_test)
    scores = {}
    for prediction_label in prediction_labels:
        latent_tmp = latent_test[~latent_test.obs[prediction_label].isnull()]
        latent_tmp = latent_tmp[latent_tmp.obs[prediction_label] != 'nan']
        latent_tmp = latent_tmp[latent_tmp.obs[prediction_label] != -99]
        sc.pp.neighbors(latent_tmp, n_neighbors=2, knn=True)
        scores[f'weighted_f1_{prediction_label}'] = Metrics.get_knn_f1_within_set(latent_tmp, prediction_label)
    return scores



def suggest_params(trial):
    dropout = trial.suggest_float('dropout', 0, 0.3, step=0.05)  # used twice
    activation = trial.suggest_categorical('activation', ['linear', 'leakyrelu'])  # used for conditional sampling
    hdim = trial.suggest_int('hdim', 100, 800, step=100)  # shared_hidden should be less than hdim
    shared_hidden = trial.suggest_int('shared_hidden', 100, min(hdim * 2, 500),
                                      step=100)  # zdim should be less than shared_hidden
    num_layers = trial.suggest_int('num_layers', 1, 3, step=1) if activation == 'leakyrelu' else 1
    tfmr_encoding_layers = trial.suggest_int('tfmr_encoding_layers', 1, 4, step=1)  # used twice
    loss_weights_kl = trial.suggest_float('loss_weights_kl', 1e-10, 1e-4, log=True)

    params = {
        'batch_size': 512,
        'learning_rate': trial.suggest_float('lr', 1e-5, 1e-3, log=True),
        'loss_weights': [0.0, 1.0, loss_weights_kl],

        'joint': {
            'activation': activation,
            'batch_norm': True,
            'dropout': dropout,
            'hdim': hdim,
            'losses': ['MSE', 'CE'],
            'num_layers': num_layers,
            'shared_hidden': [shared_hidden] * num_layers,
            'zdim': trial.suggest_int('zdim', 5, min(shared_hidden, 50), step=5),
            'c_embedding_dim': 20,
        },
        'rna': None,
        'tcr': {
            'embedding_size': trial.suggest_categorical('tfmr_embedding_size', [16, 32, 64]),
            'num_heads': trial.suggest_categorical('tfmr_num_heads', [2, 4, 8]),
            'forward_expansion': 4,
            'encoding_layers': tfmr_encoding_layers,
            'decoding_layers': tfmr_encoding_layers,  # encoding_layers is used here too
            'dropout': trial.suggest_float('tfmr_dropout', 0, 0.3, step=0.05),
        },
    }
    return params



import operator
from tcr_embedding.models.pertubation_prediction import run_scgen_cross_validation


def report_modulation_prediction(adata, model, optimization_mode_params, epoch, comet):
    degs = None
    if 'gene_set' in optimization_mode_params:
        degs = optimization_mode_params['gene_set']
    summary = run_scgen_cross_validation(adata, optimization_mode_params['column_fold'],
                                         model, optimization_mode_params['column_perturbation'],
                                         optimization_mode_params['indicator_perturbation'],
                                         degs)
    score = summary['avg_r_squared']
    if comet is not None:
        for key, value in summary.items():
            if key == 'avg_r_squared':
                comet.log_metric(key, value, step=int(epoch), epoch=epoch)
            else:
                comet.log_metrics(value, prefix=key, step=int(epoch), epoch=epoch)
    return score, operator.gt



import numpy as np
from tqdm import tqdm
from sklearn.model_selection import GroupShuffleSplit


def encode_tcr(adata, column_cdr3a, column_cdr3b, pad):
	"""
	Encodes the CDR3 alpha and CDR3 beta chain into numerical values
	:param adata: adata object
	:param column_cdr3a: column in adata.obs storing the cdr3alpha chain
	:param column_cdr3b: column in adata.obs storing the cdr3beta chain
	:param pad: int, amount of position to pad the sequence to
	:return: stores the numeric embedding to adata.obsm['alpha_seq'] and adata.obsm['beta_seq']
	"""
	aa_to_id = {'_': 0, 'A': 1, 'C': 2, 'D': 3, 'E': 4, 'F': 5, 'G': 6, 'H': 7, 'I': 8, 'K': 9, 'L': 10, 'M': 11,
				'N': 12, 'P': 13, 'Q': 14, 'R': 15, 'S': 16, 'T': 17, 'V': 18, 'W': 19, 'Y': 20, '+': 21,
				'<': 22, '>': 23}
	adata.uns['aa_to_id'] = aa_to_id
	aa_encoding(adata, read_col=column_cdr3b, label_col='beta_seq', length_col='beta_len', pad=pad, aa_to_id=aa_to_id,
				start_end_symbol=False)
	aa_encoding(adata, read_col=column_cdr3a, label_col='alpha_seq', length_col='alpha_len', pad=pad, aa_to_id=aa_to_id,
				start_end_symbol=False)


def aa_encoding(adata, read_col, ohe_col=None, label_col=None, length_col=None, pad=False, aa_to_id=None, start_end_symbol=True):
	"""
	Encoding of protein or nucleotide sequence inplace, either one-hot-encoded or as index labels and/or one-hot-encoding
	:param adata: adata file
	:param read_col: str column containing sequence
	:param ohe_col: None or str, if str column to write one-hot-encoded sequence into
	:param label_col: None or str, if str then write labels as index to this column
	:param length_col: str column None or str, if str write sequence length into this column
	:param pad: bool or int value, if int value then the sequence will be pad to this value,
				if True then pad_len will be determined by taking the longest sequence length in adata
	:param aa_to_id: None or dict, None will create a dict in this code, dict should contain {aa: index}
	:param start_end_symbol: bool, add a start '<' and end '>' symbol to each sequence
	:return:
	"""
	if label_col is None and ohe_col is None:
		raise AssertionError('Specify at least one column to write: ohe_col or label_col')

	if start_end_symbol:
		adata.obs[read_col] = '<' + adata.obs[read_col].astype('str') + '>'
		if type(pad) is not bool:
			pad += 2

	if length_col:
		adata.obs[length_col] = adata.obs[read_col].str.len()

	# Padding if specified
	if type(pad) is not bool:
		sequence_col = adata.obs[read_col].str.ljust(pad, '_')
	elif pad:
		pad_len = adata.obs[read_col].str.len().max()
		sequence_col = adata.obs[read_col].str.ljust(pad_len, '_')
	else:
		sequence_col = adata.obs[read_col]

	# tokenize each character, i.e. create list of characters
	aa_tokens = sequence_col.apply(lambda x: list(x))

	# dict containing aa name as key and token-id as value
	if aa_to_id is None:
		unique_aa_tokens = sorted(set([x for sublist in aa_tokens for x in sublist]))
		aa_to_id = {aa: id_ for id_, aa in enumerate(unique_aa_tokens)}

	# convert aa to token_id (i.e. unique integer for each aa)
	token_ids = [[aa_to_id[token] for token in aa_token] for aa_token in aa_tokens]

	# convert token_ids to one-hot
	if ohe_col is not None:
		one_hot = [np.zeros((len(aa_token), len(aa_to_id))) for aa_token in aa_tokens]
		for x_seq, token_id_seq in zip(one_hot, token_ids):
			for x, token_id in zip(x_seq, token_id_seq):
				x[token_id] = 1.0
		# adata.obs[ohe_col] = one_hot
		adata.obsm[ohe_col] = np.stack(one_hot)

	# If specified write label as index sequence
	if label_col is not None:
		token_ids = [np.array(token_id) for token_id in token_ids]
		# adata.obs[label_col] = token_ids
		adata.obsm[label_col] = np.stack(token_ids)

	adata.uns['aa_to_id'] = aa_to_id


def stratified_group_shuffle_split(df, stratify_col, group_col, val_split, random_seed=42):
	"""
	https://stackoverflow.com/a/63706321
	Split the dataset into train and test. To create a val set, execute this code twice to first split test+val and test
	and then split the test and val.

	The splitting tries to improve splitting by two properties:
	1) Stratified splitting, so the label distribution is roughly the same in both sets, e.g. antigen specificity
	2) Certain groups are only in one set, e.g. the same clonotypes are only in one set, so the model cannot peak into similar sample during training.

	If there is only one group to a label, the group is defined as training, else as test sample, the model never saw this label before.

	The outcome is not always ideal, i.e. the label distribution may not , as the labels within a group is heterogeneous (e.g. 2 cells from the same clonotype have different antigen labels)
	Also see here for the challenges: https://github.com/scikit-learn/scikit-learn/issues/12076

	:param df: pd.DataFrame containing the data to split
	:param stratify_col: str key for the column containing the classes to be stratified over all sets
	:param group_col: str key for the column containing the groups to be kept in the same set
	"""
	groups = df.groupby(stratify_col)
	all_train = []
	all_test = []
	for group_id, group in tqdm(groups):
		# if a group is already taken in test or train it must stay there
		group = group[~group[group_col].isin(all_train + all_test)]
		# if group is empty
		if group.shape[0] == 0:
			continue

		if len(group) > 1:
			train_inds, test_inds = next(
				GroupShuffleSplit(test_size=val_split, n_splits=1, random_state=random_seed).split(group, groups=group[
					group_col]))
			all_train += group.iloc[train_inds][group_col].tolist()
			all_test += group.iloc[test_inds][group_col].tolist()
		# if there is only one clonotype for this particular label
		else:
			all_train += group[group_col].tolist()

	train = df[df[group_col].isin(all_train)]
	test = df[df[group_col].isin(all_test)]

	return train, test


def group_shuffle_split(adata_tmp, group_col, val_split, random_seed=42):
	groups = adata_tmp.obs[group_col]
	splitter = GroupShuffleSplit(test_size=val_split, n_splits=5, random_state=random_seed)

	best_value = 1
	train, val = None, None
	for train_tmp, val_tmp in splitter.split(adata_tmp, groups=groups):
		split_value = abs(len(val_tmp) / len(adata_tmp) - val_split)
		if split_value < best_value:
			train = train_tmp
			val = val_tmp
			best_value = split_value

	train = adata_tmp[train]
	val = adata_tmp[val]
	return train, val



import torch
import torch.nn as nn

from tcr_embedding.models.architectures.mlp import MLP
from tcr_embedding.models.architectures.mlp_scRNA import build_mlp_encoder, build_mlp_decoder
from tcr_embedding.models.vae_base_model import VAEBaseModel


def none_model(hyperparams, hdim, xdim):
	pass


class RnaModelTorch(nn.Module):
	def __init__(self, rna_params, joint_params):
		super(RnaModelTorch, self).__init__()
		xdim = rna_params['xdim']
		hdim = joint_params['hdim']
		num_conditional_labels = joint_params['num_conditional_labels']
		cond_dim = joint_params['cond_dim']
		cond_input = joint_params['cond_input']
		zdim = joint_params['zdim']
		shared_hidden = joint_params['shared_hidden']
		activation = joint_params['activation']
		dropout = joint_params['dropout']
		batch_norm = joint_params['batch_norm']

		self.gene_encoder = build_mlp_encoder(rna_params, xdim, hdim)
		self.gene_decoder = build_mlp_decoder(rna_params, xdim, hdim)

		if num_conditional_labels > 0:
			self.cond_emb = torch.nn.Embedding(num_conditional_labels, cond_dim)
		self.cond_input = cond_input
		cond_input_dim = cond_dim if cond_input else 0

		self.shared_encoder = MLP(hdim + cond_input_dim, zdim * 2, shared_hidden, activation, 'linear',
								  dropout, batch_norm,   regularize_last_layer=False)
		self.shared_decoder = MLP(zdim + cond_dim, hdim, shared_hidden[::-1], activation, activation,
								  dropout, batch_norm, regularize_last_layer=True)

		# used for NB loss
		self.theta = torch.nn.Parameter(torch.randn(xdim))

	def forward(self, rna, tcr, tcr_len, conditional=None):
		"""
		Forward pass of autoencoder
		:param rna: torch.Tensor shape=[batch_size, num_genes]
		:param tcr: torch.Tensor shape=[batch_size, seq_len, num_seq_labels]
		:param tcr_len: torch.LongTensor shape=[batch_size] indicating how long the real unpadded length is
		:param conditional: torch.Tensor shape=[batch_size, n_cond] one-hot-encoded conditional covariates
		:return: scRNA_pred, tcr_seq_pred
		"""

		hidden_rna = self.gene_encoder(rna)

		if conditional is not None:
			cond_emb_vec = self.cond_emb(conditional)
			if self.cond_input:  # condition input flag is set
				hidden_rna = torch.cat([hidden_rna, cond_emb_vec], dim=1)  # shape=[batch_size, hdim+n_cond]

		z_ = self.shared_encoder(hidden_rna)  # shape=[batch_size, zdim*2]
		mu, logvar = z_[:, :z_.shape[1] // 2], z_[:, z_.shape[1] // 2:]  # mu.shape = logvar.shape = [batch_size, zdim]
		z = self.reparameterize(mu, logvar)  # shape=[batch_size, zdim]

		if conditional is not None:
			z_input = torch.cat([z, cond_emb_vec], dim=1)  # shape=[batch_size, hdim+cond_emb_dim]
		else:
			z_input = z  # shape=[batch_size, hdim]
		joint_dec_feature = self.shared_decoder(z_input)  # shape=[batch_size, hdim]

		rna_pred = self.gene_decoder(joint_dec_feature)  # shape=[batch_size, num_genes]
		return z, mu, logvar, rna_pred, 0

	def reparameterize(self, mu, log_var):
		"""
		https://debuggercafe.com/getting-started-with-variational-autoencoder-using-pytorch/
		:param mu: mean from the encoder's latent space
		:param log_var: log variance from the encoder's latent space
		"""
		std = torch.exp(0.5 * log_var)  # standard deviation
		eps = torch.randn_like(std)  # `randn_like` as we need the same size
		z = mu + (eps * std)  # sampling as if coming from the input space
		return z

	def predict_transcriptome(self, z_shared, conditional=None):
		"""
		Predict the transcriptome connected to an shared latent space
		:param z_shared: torch.tensor, shared latent representation
		:param conditional:
		:return: torch.tensor, transcriptome profile
		"""
		if conditional is not None:  # more efficient than doing two concatenations
			cond_emb_vec = self.cond_emb(conditional)
			z_shared = torch.cat([z_shared, cond_emb_vec], dim=-1)  # shape=[batch_size, zdim+cond_dim]
		z_shared = self.shared_decoder(z_shared)
		transcriptome_pred = self.gene_decoder(z_shared)
		return transcriptome_pred

	def get_latent_from_z(self, z):
		return z


class RnaModel(VAEBaseModel):
	def __init__(self,
				 adata,
				 params_architecture,
				 balanced_sampling='clonotype',
				 metadata=None,
				 conditional=None,
				 optimization_mode_params=None,
				 label_key=None,
				 device=None
				 ):

		super(RnaModel, self).__init__(adata, params_architecture, balanced_sampling, metadata,
									   conditional, optimization_mode_params, label_key, device)
		self.model_type = 'rna'

		self.params_rna['xdim'] = adata[0].X.shape[1]

		num_conditional_labels = 0
		cond_dim = 0
		if self.conditional is not None:
			if self.conditional in adata.obsm:
				num_conditional_labels = adata.obsm[self.conditional].shape[1]
			else:
				num_conditional_labels = len(adata.obs[self.conditional].unique())
			if 'c_embedding_dim' not in self.params_joint:
				cond_dim = 20
			else:
				cond_dim = self.params_joint['c_embedding_dim']
		self.params_joint['num_conditional_labels'] = num_conditional_labels
		self.params_joint['cond_dim'] = cond_dim
		self.params_joint['cond_input'] = conditional is not None

		self.model = RnaModelTorch(self.params_rna, self.params_joint)

	def calculate_loss(self, rna_pred, rna, tcr_pred, tcr):
		rna_loss = self.loss_weights[0] * self.loss_function_rna(rna_pred, rna)
		tcr_loss = torch.FloatTensor([0]).to(self.device)
		return rna_loss, tcr_loss

	def calculate_kld_loss(self, mu, logvar, epoch):
		kld_loss = self.loss_function_kld(mu, logvar)
		kld_loss *= self.loss_weights[2] * self.get_kl_annealing_factor(epoch)
		z = mu  # make z deterministic by using the mean
		return kld_loss, z



from tcr_embedding.evaluation.PertubationPrediction import evaluate_pertubation
import numpy as np
import scanpy as sc
import time
from sklearn.neighbors import NearestNeighbors

sc.settings.verbosity = 0


def predict_pertubation(latent_train, latent_val, model, column_perturbation, indicator_perturbation, var_names,
                        col_type, return_latent=False):
    """
    Predict the effect of pertubation on transcriptome level
    :param latent_train: adata object containing the latent spaces of the training dataset
    :param latent_val: adata object containing the latent spaces of the validaiton dataset
    :param model: model to predict transcritome from latent space
    :param column_perturbation: str, column in the adata objects indicating the pertubation
    :param indicator_perturbation: str, value for the 'pre' state of the perturbation
    :param var_names: list containing the gene names
    :return: adata object, containing the predicted transcriptome profile after perturbation
    """
    latent_val_pre = latent_val[latent_val.obs[column_perturbation] == indicator_perturbation]

    cts_pre = latent_train[latent_train.obs[column_perturbation] == indicator_perturbation].obs[col_type].unique()
    cts_post = latent_train[latent_train.obs[column_perturbation] != indicator_perturbation].obs[col_type].unique()
    cts_both = [ct for ct in cts_pre if ct in cts_post]

    centers = latent_train[latent_train.obs[col_type].isin(cts_both) &
                           (latent_train.obs[column_perturbation] == indicator_perturbation)]
    centers = [centers[centers.obs[col_type] == ct_group].X.mean(axis=0) for ct_group in cts_both]

    center_pre = latent_val.X.mean(axis=0)

    nn = NearestNeighbors(1, metric='euclidean', algorithm='brute').fit(centers)
    _, idx = nn.kneighbors(center_pre.reshape(1, -1))
    idx = idx[0][0]

    center_post = latent_train[(latent_train.obs[col_type] == cts_both[idx]) &
                               (latent_train.obs[column_perturbation] != indicator_perturbation)].X.mean(axis=0)
    delta = center_post - centers[idx]

    ad_pred = sc.AnnData(latent_val_pre.X + delta, obs=latent_val.obs.copy())
    if return_latent:
        return ad_pred

    ad_pred = model.predict_rna_from_latent(ad_pred, metadata=ad_pred.obs.columns)
    ad_pred.var_names = var_names
    return ad_pred


def get_delta(adata_latent, column_perturbation, indicator_perturbation):
    """
    Calculate the difference vector between pre and post perturbation in the latent space
    :param adata_latent: adata oject, containing the latent space representation of the training data
    :param column_perturbation: str, column in the adata object indicating the pertubation
    :param indicator_perturbation: str, value for the 'pre' state of the profile after perturbation
    :return: numpy array, difference vectore
    """
    mask_pre = adata_latent.obs[column_perturbation] == indicator_perturbation
    latent_pre = adata_latent[mask_pre]
    latent_post = adata_latent[~mask_pre]
    avg_pre = np.mean(latent_pre.X, axis=0)
    avg_post = np.mean(latent_post.X, axis=0)
    delta = avg_post - avg_pre
    return delta


def run_scgen_cross_validation(adata, column_fold, model, column_perturbation, indicator_perturbation, degs=None):
    """
    Runs perturbation prediction over a specified fold column and evaluates the results
    :param adata: adata object, of the raw data
    :param column_fold: str, indicating the column over which to cross validate
    :param model: model used for latent space generation
    :param column_perturbation: str, column in the adata object indicating the pertubation
    :param indicator_perturbation: str, value for the 'pre' state of the profile after perturbation
    :return: dict, summary over performance on the different splits and aggregation
    """
    tic = time.time()
    latent_full = model.get_latent(adata, metadata=[column_fold, column_perturbation])

    summary_performance = {}
    rs_squared = []

    cats_pre = adata[adata.obs[column_perturbation] == indicator_perturbation].obs[column_fold].value_counts()
    cats_pre = [cat for cat, count in cats_pre.items() if count > 5]
    cats_post = adata[adata.obs[column_perturbation] != indicator_perturbation].obs[column_fold].value_counts()
    cats_post = [cat for cat, count in cats_post.items() if count > 5]
    cats_both = [cat for cat in cats_pre if cat in cats_post]

    centers_pre = [latent_full[(latent_full.obs[column_fold] == cat_group) &
                               (latent_full.obs[column_perturbation] == indicator_perturbation)].X.mean(axis=0)
                   for cat_group in cats_both]
    centers_post = [latent_full[(latent_full.obs[column_fold] == cat_group) &
                                (latent_full.obs[column_perturbation] != indicator_perturbation)].X.mean(axis=0)
                    for cat_group in cats_both]

    for i, fold in enumerate(cats_both):
        cats_tmp = cats_both.copy()
        cats_tmp.pop(i)
        centers_pre_tmp = centers_pre.copy()
        centers_pre_tmp.pop(i)
        centers_post_tmp = centers_post.copy()
        centers_post_tmp.pop(i)

        mask_train = latent_full.obs[column_fold] != fold
        latent_val = latent_full[~mask_train]
        latent_val_pre = latent_val[latent_val.obs[column_perturbation] == indicator_perturbation]

        center_pre = latent_val_pre.X.mean(axis=0)
        nn = NearestNeighbors(1, metric='euclidean', algorithm='brute').fit(centers_pre_tmp)
        _, idx = nn.kneighbors(center_pre.reshape(1, -1))
        idx = idx[0][0]

        delta = centers_post_tmp[idx] - centers_pre_tmp[idx]

        ad_pred = sc.AnnData(latent_val_pre.X + delta, obs=latent_val_pre.obs.copy())
        ad_pred = model.predict_rna_from_latent(ad_pred, metadata=ad_pred.obs.columns)
        ad_pred.var_names = adata.var_names

        data_ct = adata[adata.obs[column_fold] == fold].copy()
        score = evaluate_pertubation(data_ct.copy(), ad_pred, column_fold, column_perturbation,
                                     indicator=indicator_perturbation, gene_set=degs)

        for key, value in score.items():
            summary_performance[f'{fold}_key'] = value
        rs_squared.append(score['degs']['r_squared'])

    summary_performance['avg_r_squared'] = sum(rs_squared) / len(rs_squared)
    toc = time.time()
    #print(summary_performance['avg_r_squared'])
    #print(toc - tic)
    return summary_performance



import scanpy as sc
import tcr_embedding.evaluation.Metrics as Metrics


def evaluate_pertubation(data_val, prediction, per_column, pertubation, indicator='pre', gene_set=None):
    """
    Evaluate pertubation prediction by Pearson squared for all genes and top100, overall and per classes
    :param data_val: adata object containing the ground truth validation data
    :param prediction: adata object containing the predicted latent space
    :param per_column: column from which the classes are used
    :param pertubation: column of pertubation
    :param indicator: value in pertubation indicating pre state
    :param gene_set: all metrics will be measured on this gene subset
    :return: dictionary containing a summary of the performance
    """
    ground_truth = data_val[data_val.obs[pertubation] != indicator].copy()

    # evaluate over all genes and cell types
    summary = {'all_genes': Metrics.get_square_pearson(ground_truth, prediction)}

    # evaluate by column
    if per_column is not None:
        groups = data_val.obs[per_column].unique()
        summary[f'per_{per_column}'] = evaluate_per_column(ground_truth, prediction, per_column, groups)

    # evaluate over top 100 genes
    if gene_set is None:
        return summary

    ground_truth = ground_truth[:, gene_set]
    prediction = prediction[:, gene_set]

    # evaluate over top 100 genes per cell type
    summary['degs'] = Metrics.get_square_pearson(ground_truth, prediction)
    # evaluate by column
    if per_column is not None:
        groups = data_val.obs[per_column].unique()
        summary[f'per_{per_column}_degs'] = evaluate_per_column(ground_truth, prediction, per_column, groups)
    return summary


def evaluate_per_column(ground_truth, prediction, per_column, groups):
    """
    Evaluate scores for subcategories of the cells
    :param ground_truth: adata object containing the real transcriptomic data
    :param prediction: adata object containing the predicted transcriptomic data
    :param per_column: column for grouping
    :param groups: possible entries in this column
    :return: dict {group_name: performance_summary}
    """
    sub_summary = {}
    for entry in groups:
        if entry not in ground_truth.obs[per_column].unique() or entry not in prediction.obs[per_column].unique():
            continue
        sub_prediction = prediction[prediction.obs[per_column] == entry]
        sub_ground_truth = ground_truth[ground_truth.obs[per_column] == entry]
        sub_summary[entry] = Metrics.get_square_pearson(sub_ground_truth, sub_prediction)
    return sub_summary



# https://github.com/theislab/multigrate
import torch

class NB(torch.nn.Module):
    """
    Yang Comment: Usage in forward:
    x : Ground truth
    mu: Prediction
    theta: Another trainable parameter with shape=[xdim(number of count variables)],
           simply initialize a nn.Parameter(torch.randn(xdim)) in the model

    Be careful, we need the negative of the returned value: loss = -NBLoss
    """
    def __init__(self, eps=1e-8, reduction='mean'):
        super(NB, self).__init__()
        self.eps = eps
        self.reduction = reduction

    def forward(self, x: torch.Tensor, mu: torch.Tensor, theta: torch.Tensor):

        if theta.ndimension() == 1:
            theta = theta.view(1, theta.size(0))

        log_theta_mu_eps = torch.log(theta + mu + self.eps)
        res = (
            theta * (torch.log(theta + self.eps) - log_theta_mu_eps)
            + x * (torch.log(mu + self.eps) - log_theta_mu_eps)
            + torch.lgamma(x + theta)
            - torch.lgamma(theta)
            - torch.lgamma(x + 1)
        )

        if self.reduction == 'mean':
            res = torch.mean(res)
        elif self.reduction == 'sum':
            res = torch.sum(res)
        else:
            raise NotImplementedError(f'reduction method {self.reduction} is not implemented.')

        return res

"""
Code Snippet from vae base
prediction = F.softmax(prediction, dim=-1)
size_factor_view = size_factor.unsqueeze(1).expand(-1, prediction.shape[1]).to(prediction.device)
dec_mean = prediction * size_factor_view
dispersion = self.model.theta.T
dispersion = torch.exp(dispersion)
loss = - scRNA_criterion(prediction, dec_mean,
                                          dispersion) * 1000  # TODO A Test, maybe delete afterwards
"""


# A Variational Information Bottleneck Approach to Multi-Omics Data Integration
import torch
import torch.nn as nn

from tcr_embedding.models.architectures.transformer import TransformerEncoder, TransformerDecoder
from tcr_embedding.models.architectures.mlp import MLP
from tcr_embedding.models.architectures.mlp_scRNA import build_mlp_encoder, build_mlp_decoder
from tcr_embedding.models.vae_base_model import VAEBaseModel


class PoEModelTorch(nn.Module):
    def __init__(self, tcr_params, rna_params, joint_params):
        super(PoEModelTorch, self).__init__()
        xdim = rna_params['xdim']
        hdim = joint_params['hdim']
        num_conditional_labels = joint_params['num_conditional_labels']
        cond_dim = joint_params['cond_dim']
        cond_input = joint_params['cond_input']
        zdim = joint_params['zdim']
        shared_hidden = joint_params['shared_hidden']
        activation = joint_params['activation']
        dropout = joint_params['dropout']
        batch_norm = joint_params['batch_norm']

        num_seq_labels = tcr_params['num_seq_labels']

        self.alpha_encoder = TransformerEncoder(tcr_params, hdim // 2, num_seq_labels)
        self.alpha_decoder = TransformerDecoder(tcr_params, hdim, num_seq_labels)

        self.beta_encoder = TransformerEncoder(tcr_params, hdim // 2, num_seq_labels)
        self.beta_decoder = TransformerDecoder(tcr_params, hdim, num_seq_labels)

        self.rna_encoder = build_mlp_encoder(rna_params, xdim, hdim)
        self.rna_decoder = build_mlp_decoder(rna_params, xdim, hdim)

        if cond_dim > 0:
            self.cond_emb = torch.nn.Embedding(num_conditional_labels, cond_dim)
        self.cond_input = cond_input
        cond_input_dim = cond_dim if cond_input else 0

        self.tcr_vae_encoder = MLP(hdim + cond_input_dim, zdim * 2, shared_hidden, activation, 'linear', dropout,
                                   batch_norm, regularize_last_layer=False)
        self.tcr_vae_decoder = MLP(zdim + cond_dim, hdim, shared_hidden[::-1], activation, activation, dropout,
                                   batch_norm, regularize_last_layer=True)

        self.rna_vae_encoder = MLP(hdim + cond_input_dim, zdim * 2, shared_hidden, activation, 'linear', dropout,
                                   batch_norm, regularize_last_layer=False)
        self.rna_vae_decoder = MLP(zdim + cond_dim, hdim, shared_hidden[::-1], activation, activation, dropout,
                                   batch_norm, regularize_last_layer=True)

        # used for NB loss
        self.theta = torch.nn.Parameter(torch.randn(xdim))

    def forward(self, rna, tcr, tcr_len, conditional=None):
        """
		Forward pass of autoencoder
		:param rna: torch.Tensor shape=[batch_size, num_genes]
		:param tcr: torch.Tensor shape=[batch_size, seq_len, feature_dim]
		:param tcr_len: torch.LongTensor shape=[batch_size] indicating how long the real unpadded length is
		:param conditional: torch.Tensor shape=[batch_size, n_cond] one-hot-encoded conditional covariates
		:return:
			z: list of sampled latent variable zs. z = [z_rna, z_tcr, z_joint]
			mu: list of predicted means mu. mu = [mu_rna, mu_tcr, mu_joint]
			logvar: list of predicted logvars. logvar = [logvar_rna, logvar_tcr, logvar_joint]
			rna_pred: list of reconstructed rna. rna_pred = [rna_pred using z_rna, rna_pred using z_joint]
			tcr_pred: list of reconstructed tcr. tcr_pred = [tcr_pred using z_tcr, tcr_pred using z_joint]
		"""
        if conditional is not None:
            cond_emb_vec = self.cond_emb(conditional)
        # Encode TCR
        alpha_seq = tcr[:, :tcr.shape[1] // 2]
        alpha_len = tcr_len[:, 0]

        beta_seq = tcr[:, tcr.shape[1] // 2:]
        beta_len = tcr_len[:, 1]

        h_alpha = self.alpha_encoder(alpha_seq, alpha_len)  # shape=[batch_size, hdim//2]
        h_beta = self.beta_encoder(beta_seq, beta_len)  # shape=[batch_size, hdim//2]
        h_tcr = torch.cat([h_alpha, h_beta], dim=-1)  # shape=[batch_size, hdim]
        if conditional is not None and self.cond_input:
            h_tcr = torch.cat([h_tcr, cond_emb_vec], dim=1)  # shape=[batch_size, hdim+n_cond]

        # Encode RNA
        h_rna = self.rna_encoder(rna)  # shape=[batch_size, hdim]
        if conditional is not None and self.cond_input:
            h_rna = torch.cat([h_rna, cond_emb_vec], dim=1)  # shape=[batch_size, hdim+n_cond]

        # Predict Latent space
        z_rna_ = self.rna_vae_encoder(h_rna)  # shape=[batch_size, zdim*2]
        mu_rna, logvar_rna = z_rna_[:, :z_rna_.shape[1] // 2], z_rna_[:, z_rna_.shape[1] // 2:]
        z_rna = self.reparameterize(mu_rna, logvar_rna)  # shape=[batch_size, zdim]

        z_tcr_ = self.tcr_vae_encoder(h_tcr)  # shape=[batch_size, zdim*2]
        mu_tcr, logvar_tcr = z_tcr_[:, :z_tcr_.shape[1] // 2], z_tcr_[:, z_tcr_.shape[1] // 2:]
        z_tcr = self.reparameterize(mu_tcr, logvar_tcr)  # shape=[batch_size, zdim]

        # Predict joint latent space using PoE
        mu_joint, logvar_joint = self.product_of_experts(mu_rna, mu_tcr, logvar_rna, logvar_tcr)
        z_joint = self.reparameterize(mu_joint, logvar_joint)

        z = [z_rna, z_tcr, z_joint]
        mu = [mu_rna, mu_tcr, mu_joint]
        logvar = [logvar_rna, logvar_tcr, logvar_joint]

        # Reconstruction
        rna_pred = []
        for z_ in [z_rna, z_joint]:
            if conditional is not None:
                z_ = torch.cat([z_, cond_emb_vec], dim=1)  # shape=[batch_size, hdim+n_cond]
            f_rna = self.rna_vae_decoder(z_)
            rna_pred.append(self.rna_decoder(f_rna))
        tcr_pred = []
        for z_ in [z_tcr, z_joint]:
            if conditional is not None:
                z_ = torch.cat([z_, cond_emb_vec], dim=1)  # shape=[batch_size, hdim+n_cond]
            f_tcr = self.tcr_vae_decoder(z_)
            alpha_pred = self.alpha_decoder(f_tcr, alpha_seq)
            beta_pred = self.beta_decoder(f_tcr, beta_seq)

            tcr_pred.append(torch.cat([alpha_pred, beta_pred], dim=1))

        return z, mu, logvar, rna_pred, tcr_pred

    def reparameterize(self, mu, log_var):
        """
        https://debuggercafe.com/getting-started-with-variational-autoencoder-using-pytorch/
        :param mu: mean from the encoder's latent space
        :param log_var: log variance from the encoder's latent space
        """
        std = torch.exp(0.5 * log_var)  # standard deviation
        eps = torch.randn_like(std)  # `randn_like` as we need the same size
        z = mu + (eps * std)  # sampling as if coming from the input space
        return z

    def product_of_experts(self, mu_rna, mu_tcr, logvar_rna, logvar_tcr):
        # formula: var_joint = inv(inv(var_prior) + sum(inv(var_modalities)))
        logvar_joint = 1.0 / torch.exp(logvar_rna) + 1.0 / torch.exp(
            logvar_tcr) + 1.0  # sum up all inverse vars, logvars first needs to be converted to var, last 1.0 is coming from the prior
        logvar_joint = torch.log(1.0 / logvar_joint)  # inverse and convert to logvar

        # formula: mu_joint = (mu_prior*inv(var_prior) + sum(mu_modalities*inv(var_modalities))) * var_joint, where mu_prior = 0.0
        mu_joint = mu_rna * (1.0 / torch.exp(logvar_rna)) + mu_tcr * (1.0 / torch.exp(logvar_tcr))
        mu_joint = mu_joint * torch.exp(logvar_joint)

        return mu_joint, logvar_joint

    def predict_transcriptome(self, z_shared, conditional=None):
        """
        Predict the transcriptome connected to an shared latent space
        :param z_shared: torch.tensor, shared latent representation
        :param conditional:
        :return: torch.tensor, transcriptome profile
        """
        if conditional is not None:  # more efficient than doing two concatenations
            cond_emb_vec = self.cond_emb(conditional)
            z_shared = torch.cat([z_shared, cond_emb_vec], dim=-1)  # shape=[batch_size, zdim+cond_dim]
        transcriptome_pred = self.rna_vae_decoder(z_shared)
        transcriptome_pred = self.rna_decoder(transcriptome_pred)
        return transcriptome_pred

    def get_latent_from_z(self, z):
        return z[2]


class PoEModel(VAEBaseModel):
    def __init__(self,
                 adata,
                 params_architecture,
                 balanced_sampling='clonotype',
                 metadata=None,
                 conditional=None,
                 optimization_mode_params=None,
                 label_key=None,
                 device=None
                 ):
        super(PoEModel, self).__init__(adata, params_architecture, balanced_sampling, metadata,
                                       conditional, optimization_mode_params, label_key, device)

        self.model_type = 'poe'

        self.params_tcr['max_tcr_length'] = adata.obsm['alpha_seq'].shape[1]
        self.params_tcr['num_seq_labels'] = len(self.aa_to_id)

        self.params_rna['xdim'] = adata[0].X.shape[1]

        num_conditional_labels = 0
        cond_dim = 0
        if self.conditional is not None:
            if self.conditional in adata.obsm:
                num_conditional_labels = adata.obsm[self.conditional].shape[1]
            else:
                num_conditional_labels = len(adata.obs[self.conditional].unique())
            if 'c_embedding_dim' not in self.params_joint:
                cond_dim = 20
            else:
                cond_dim = self.params_joint['c_embedding_dim']
        self.params_joint['num_conditional_labels'] = num_conditional_labels
        self.params_joint['cond_dim'] = cond_dim
        self.params_joint['cond_input'] = conditional is not None

        self.model = PoEModelTorch(self.params_tcr, self.params_rna, self.params_joint)

    def calculate_loss(self, rna_pred, rna, tcr_pred, tcr):
        # Evaluate same-modality and joint reconstruction
        rna_loss = self.loss_function_rna(rna_pred[0], rna) + self.loss_function_rna(rna_pred[1], rna)
        rna_loss *= 0.5 * self.loss_weights[0]

        # For GRU and Transformer, as they don't predict start token for alpha and beta chain, so -2
        if tcr_pred[0].shape[1] == tcr.shape[1] - 2:
            mask = torch.ones_like(tcr).bool()
            mask[:, [0, mask.shape[1] // 2]] = False
            tcr_loss = (self.loss_function_tcr(tcr_pred[0].flatten(end_dim=1), tcr[mask].flatten()) +
                        self.loss_function_tcr(tcr_pred[1].flatten(end_dim=1), tcr[mask].flatten()))
            tcr_loss *= 0.5 * self.loss_weights[1]
        else:  # For CNN, as it predicts start token
            tcr_loss = (self.loss_function_tcr(tcr_pred[0].flatten(end_dim=1), tcr.flatten()) +
                        self.loss_function_tcr(tcr_pred[1].flatten(end_dim=1), tcr.flatten()))
            tcr_loss *= 0.5 * self.loss_weights[1]
        return rna_loss, tcr_loss

    def calculate_kld_loss(self, mu, logvar, epoch):
        """
		Calculate the kld loss for rna, tcr, and joint latent space (optional between rna and joint)
		:param mu:
		:param logvar:
		:param epoch:
		:return:
		"""
        kld_loss = (self.loss_function_kld(mu[0], logvar[0])
                    + self.loss_function_kld(mu[1], logvar[1])
                    + self.loss_function_kld(mu[2], logvar[2]))
        kld_loss *= 1.0 / 3.0 * self.loss_weights[2] * self.get_kl_annealing_factor(epoch)

        # possible constrain on joint space to resemble more the TCR space
        if len(self.loss_weights) == 4:
            kld_rna_joint = self.loss_function_kld(mu[0], logvar[0], mu[2], logvar[2])
            kld_rna_joint = self.loss_weights[3] * kld_rna_joint
            kld_loss += kld_rna_joint
        z = mu[2]  # use joint latent variable for further downstream tasks
        return kld_loss, z



def suggest_params(trial):
    dropout = trial.suggest_float('dropout', 0, 0.3, step=0.05)  # used twice
    activation = trial.suggest_categorical('activation', ['linear', 'leakyrelu'])  # used for conditional sampling
    rna_hidden = trial.suggest_int('rna_hidden', 500, 2000, step=250)  # hdim should be less than rna_hidden
    hdim = trial.suggest_int('hdim', 100, min(rna_hidden, 800), step=100)  # shared_hidden should be less than hdim
    shared_hidden = trial.suggest_int('shared_hidden', 100, min(hdim * 2, 500),
                                      step=100)  # zdim should be less than shared_hidden
    num_layers = trial.suggest_int('num_layers', 1, 3, step=1) if activation == 'leakyrelu' else 1
    rna_num_layers = trial.suggest_int('rna_num_layers', 1, 3, step=1)
    loss_weights_kl = trial.suggest_float('loss_weights_kl', 1e-10, 1e-4, log=True)

    params = {
        'batch_size': 512,
        'learning_rate': trial.suggest_float('lr', 1e-5, 1e-3, log=True),
        'loss_weights': [1.0, 0.0, loss_weights_kl],

        'joint': {
            'activation': activation,
            'batch_norm': True,
            'dropout': dropout,
            'hdim': hdim,
            'losses': ['MSE', 'CE'],
            'num_layers': num_layers,
            'shared_hidden': [shared_hidden] * num_layers,
            'zdim': trial.suggest_int('zdim', 5, min(shared_hidden, 50), step=5),
            'c_embedding_dim': 20,
        },
        'rna': {
            'activation': 'leakyrelu',
            'batch_norm': True,
            'dropout': dropout,
            'gene_hidden': [rna_hidden] * rna_num_layers,
            'num_layers': rna_num_layers,
            'output_activation': 'linear'
        },
        'seq_model_hyperparams': None,
    }
    return params



import torch
import numpy as np
from scipy import sparse


class JointDataset(torch.utils.data.Dataset):
	def __init__(
			self,
			rna_data,
			tcr_data,
			tcr_length,
			metadata,
			labels=None,
			conditional=None
	):
		"""
		:param rna_data: list of gene expressions, where each element is a numpy or sparse matrix of one dataset
		:param tcr_data: list of seq_data, where each element is a seq_list of one dataset
		:param tcr_length: list of non-padded sequence length, needed for many architectures to mask the padding out
		:param metadata: list of metadata
		:param labels: list of labels
		:param conditional: list of conditionales
		"""
		self.metadata = metadata.tolist()
		self.tcr_length = torch.LongTensor(tcr_length)

		if conditional is not None:
			# Reduce the one-hot-encoding back to labels
			self.conditional = torch.LongTensor(conditional.argmax(1))
		# LongTensor since it is going to be embedded
		else:
			self.conditional = None

		self.rna_data = self.create_tensor(rna_data)
		# self.size_factors = self.rna_data.sum(1)

		self.tcr_data = torch.LongTensor(tcr_data)

		if labels is not None:
			self.labels = torch.LongTensor(self.labels)
		else:
			self.labels = None

	def create_tensor(self, x):
		if sparse.issparse(x):
			x = x.todense()
			return torch.FloatTensor(x)
		else:
			return torch.FloatTensor(x)

	def __len__(self):
		return len(self.rna_data)

	def __getitem__(self, idx):
		if self.labels is None:
			if self.conditional is None:
				return self.rna_data[idx], self.tcr_data[idx], self.tcr_length[idx], \
					   self.metadata[idx], False, False
			else:
				return self.rna_data[idx], self.tcr_data[idx], self.tcr_length[idx], \
					   self.metadata[idx], False, self.conditional[idx]
		else:
			if self.conditional is None:
				return self.rna_data[idx], self.tcr_datas[idx], self.tcr_lenght[idx], \
					   self.metadata[idx], self.labels[idx], False
			else:
				return self.rna_data[idx], self.tcr_datas[idx], self.tcr_lenght[idx], \
					   self.metadata[idx], self.labels[idx], self.conditional[idx]


class DeepTCRDataset(torch.utils.data.Dataset):
	def __init__(
			self,
			alpha_seq,
			beta_seq,
			vdj_dict,
			metadata
	):
		"""
		:param alpha_seq:
		:param beta_seq:
		:param vdj_dict:
		:param metadata: list of metadata
		"""
		self.metadata = np.concatenate(metadata, axis=0).tolist()

		# Concatenate datasets to be able to shuffle data through

		self.alpha_seq = np.concatenate(alpha_seq)
		self.alpha_seq = torch.LongTensor(self.alpha_seq)

		self.beta_seq = np.concatenate(beta_seq)
		self.beta_seq = torch.LongTensor(self.beta_seq)

		v_alpha = torch.LongTensor(vdj_dict['v_alpha'])
		j_alpha = torch.LongTensor(vdj_dict['j_alpha'])

		v_beta = torch.LongTensor(vdj_dict['v_beta'])
		d_beta = torch.LongTensor(vdj_dict['d_beta'])
		j_beta = torch.LongTensor(vdj_dict['j_beta'])

		self.vdj = torch.stack([v_alpha, j_alpha, v_beta, d_beta, j_beta], dim=1)

	def __len__(self):
		return len(self.alpha_seq)

	def __getitem__(self, idx):
		return self.alpha_seq[idx], self.beta_seq[idx], self.vdj[idx], self.metadata[idx]



import math
import torch
import torch.nn as nn


# from https://pytorch.org/tutorials/beginner/transformer_tutorial.html
class TrigonometricPositionalEncoding(nn.Module):
    """
        This trigonometric positional embedding was taken from:
        Title: Sequence-to-Sequence Modeling with nn.Transformer and TorchText
        Date: 17th March 2021
        Availability: https://pytorch.org/tutorials/beginner/transformer_tutorial.html
    """
    def __init__(self, embedding_dim, dropout, max_len):
        super(TrigonometricPositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)
        pe = torch.zeros(max_len, embedding_dim)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, embedding_dim, 2).float() * (-math.log(10000.0) / embedding_dim))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer('pe', pe)

    def forward(self, x):
        x = x + self.pe[:x.size(0), :]
        return self.dropout(x)


class TransformerEncoder(nn.Module):
    def __init__(self, params, hdim, num_seq_labels):
        """
        :param params: hyperparameters as dict
        :param hdim: output feature dimension
        :param num_seq_labels: number of aa labels, input dim
        """
        super(TransformerEncoder, self).__init__()
        self.params = params

        self.num_seq_labels = num_seq_labels

        self.embedding = nn.Embedding(num_seq_labels, params['embedding_size'], padding_idx=0)
        self.positional_encoding = TrigonometricPositionalEncoding(params['embedding_size'],
                                                                   params['dropout'],
                                                                   params['max_tcr_length'])

        encoding_layers = nn.TransformerEncoderLayer(params['embedding_size'],
                                                     params['num_heads'],
                                                     params['embedding_size'] * params['forward_expansion'],
                                                     params['dropout'])
        self.transformer_encoder = nn.TransformerEncoder(encoding_layers, params['encoding_layers'])

        self.fc_reduction = nn.Linear(params['max_tcr_length'] * params['embedding_size'], hdim)

    def forward(self, x, tcr_len):
        x = self.embedding(x) * math.sqrt(self.num_seq_labels)
        x = x.transpose(0, 1)
        x = x + self.positional_encoding(x)
        x = self.transformer_encoder(x)
        x = x.transpose(0, 1)
        x = x.flatten(1)
        x = self.fc_reduction(x)
        return x


class TransformerDecoder(nn.Module):
    def __init__(self, params, hdim, num_seq_labels):
        """

        :param params: hyperparameters as dict
        :param hdim: input feature dimension
        :param num_seq_labels: number of aa labels, output dim
        """
        super(TransformerDecoder, self).__init__()
        self.params = params
        self.hdim = hdim
        self.num_seq_labels = num_seq_labels

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        self.fc_upsample = nn.Linear(hdim, self.params['max_tcr_length'] * params['embedding_size'])
        # the embedding size remains constant over all layers

        self.embedding = nn.Embedding(num_seq_labels, params['embedding_size'], padding_idx=0)
        self.positional_encoding = TrigonometricPositionalEncoding(params['embedding_size'],
                                                                   params['dropout'],
                                                                   self.params['max_tcr_length'])

        decoding_layers = nn.TransformerDecoderLayer(params['embedding_size'],
                                                     params['num_heads'],
                                                     params['embedding_size'] * params['forward_expansion'],
                                                     params['dropout'])
        self.transformer_decoder = nn.TransformerDecoder(decoding_layers, params['decoding_layers'])

        self.fc_out = nn.Linear(params['embedding_size'], num_seq_labels)

    def forward(self, hidden_state, target_sequence):
        """
        Forward pass of the Decoder module
        :param hidden_state: joint hidden state of the VAE
        :param target_sequence: Ground truth output
        :return:
        """
        hidden_state = self.fc_upsample(hidden_state)
        shape = (hidden_state.shape[0], self.params['max_tcr_length'], self.params['embedding_size'])
        hidden_state = torch.reshape(hidden_state, shape)

        hidden_state = hidden_state.transpose(0, 1)

        target_sequence = target_sequence[:, :-1]
        target_sequence = target_sequence.transpose(0, 1)

        target_sequence = self.embedding(target_sequence) * math.sqrt(self.num_seq_labels)
        target_sequence = target_sequence + self.positional_encoding(target_sequence)
        try:
            target_mask = nn.Transformer.generate_square_subsequent_mask(None, target_sequence.shape[0]).to(self.device)
        except:  # new version don't need the None
            target_mask = nn.Transformer.generate_square_subsequent_mask(target_sequence.shape[0]).to(self.device)
        x = self.transformer_decoder(target_sequence, hidden_state, tgt_mask=target_mask)
        x = self.fc_out(x)
        x = x.transpose(0, 1)
        return x



import torch
import torch.nn as nn
import math


class CNNEncoder(nn.Module):
	def __init__(self, params, hdim, num_seq_labels, use_output_layer=True, use_embedding_matrix=False):
		"""
		Based on paper: DeepTCR is a deep learning framework for revealing sequence concepts within T-cell repertoires
		:param params: hyperparameters as dict
		:param hdim: int, output feature dimension
		:param num_seq_labels: int, number of aa labels, input dim (feature dim)
		"""
		super(CNNEncoder, self).__init__()
		input_len = params['max_tcr_length']

		params_enc = params['encoder']
		params_enc['num_features'] = params['num_features']
		params_enc['dropout'] = params['dropout']
		params_enc['batch_norm'] = params['batch_norm']
		params_enc['embedding_dim'] = params['embedding_dim']
		params = params_enc

		if type(params['kernel']) == int:
			kernel = [params['kernel']] * params['num_layers']
		elif type(params['kernel']) == list:
			kernel = params['kernel']
		else:
			raise ValueError()

		if type(params['num_features']) == int:
			num_features = [params['num_features']] * params['num_layers']
		elif type(params['num_features']) == list:
			num_features = params['num_features']
		else:
			raise ValueError()

		if type(params['stride']) == int:
			stride = [params['stride']] * params['num_layers']
		elif type(params['stride']) == list:
			stride = params['stride']
		else:
			raise ValueError()

		assert len(stride) == len(kernel)
		assert len(kernel) == len(num_features)

		self.use_embedding_matrix = use_embedding_matrix
		self.num_seq_labels = num_seq_labels
		if use_embedding_matrix:
			self.one_hot = nn.functional.one_hot
			self.embedding = nn.Parameter(torch.randn(num_seq_labels, params['embedding_dim']))
		else:
			self.embedding = nn.Embedding(num_embeddings=num_seq_labels, embedding_dim=params['embedding_dim'], padding_idx=0)

		input_features = [params['embedding_dim']] + num_features[:-1]
		output_len = input_len  # Sequence length after all convolution blocks
		cnn_blocks = []
		for input_, output_, kernel_, stride_ in zip(input_features, num_features, kernel, stride):
			padding = (kernel_ - 1) // 2
			output_len = math.floor((output_len + (2 * padding) - kernel_) / stride_ + 1)
			block = []
			block.append(nn.Conv1d(input_, output_, kernel_, stride_, padding=padding, bias=False if params['batch_norm'] else True))
			if params['batch_norm']:
				block.append(nn.BatchNorm1d(output_))
			if params['activation'] != 'linear':
				block.append(self._activation(params['activation']))
			if params['dropout'] is not None and params['dropout'] != 0.0:
				block.append(nn.Dropout(params['dropout']))
			cnn_blocks.append(nn.Sequential(*block))

		self.cnn_blocks = nn.Sequential(*cnn_blocks)

		self.use_output_layer = use_output_layer
		self.output_dim = num_features[-1] * output_len
		if use_output_layer:
			self.output_layer = nn.Linear(num_features[-1] * output_len, hdim)

	def forward(self, tcr_seq, tcr_len):
		if self.use_embedding_matrix:
			x = torch.matmul(self.one_hot(tcr_seq, self.num_seq_labels).float(), self.embedding)
		else:
			x = self.embedding(tcr_seq)  # shape=[batch, sequence, feature]
		x = x.permute(0, 2, 1)  # shape=[batch, feature, sequence]
		x = self.cnn_blocks(x)  # shape=[batch, feature, sequence]

		x = x.flatten(1)
		if self.use_output_layer:
			x = self.output_layer(x)
		return x

	def _activation(self, name):
		if name == 'relu':
			return nn.ReLU(inplace=True)
		elif name == 'leakyrelu':
			return nn.LeakyReLU(0.2, inplace=True)
		elif name == 'sigmoid':
			return nn.Sigmoid()
		elif name == 'softmax':
			return nn.Softmax()
		else:
			raise NotImplementedError(f'activation function {name} is not implemented.')


class CNNDecoder(nn.Module):
	def __init__(self, params, hdim, num_seq_labels, use_embedding_matrix=False):
		"""
		Based on paper: DeepTCR is a deep learning framework for revealing sequence concepts within T-cell repertoires
		:param params: hyperparameters as dict
		:param hdim: input feature dimension
		:param num_seq_labels: number of aa labels, output dim
		"""
		super(CNNDecoder, self).__init__()

		self.max_len = params['max_tcr_length']
		params_dec = params['decoder']
		params_dec['num_features'] = [params_dec['initial_feature']] + params['num_features'][1:][::-1]
		params_dec['dropout'] = params['dropout']
		params_dec['batch_norm'] = params['batch_norm']
		params_dec['embedding_dim'] = params['embedding_dim']
		params = params_dec

		if type(params['kernel']) == int:
			kernel = [params['kernel']] * (params['num_layers'] - 1)  # last dim is determined automatically
		elif type(params['kernel']) == list:
			kernel = params['kernel']
		else:
			raise ValueError()

		if type(params['num_features']) == int:
			num_features = [params['num_features']] * params['num_layers']
		elif type(params['num_features']) == list:
			num_features = params['num_features']
		else:
			raise ValueError()

		if type(params['stride']) == int:
			stride = [params['stride']] * (params['num_layers'] - 1)  # last dim is determined automatically
		elif type(params['stride']) == list:
			stride = params['stride']
		else:
			raise ValueError()

		assert len(stride) == len(kernel)
		assert len(kernel) == len(num_features) - 1

		# out = (in - 1) * stride + kernel_size
		self.initial_len = params['initial_len']
		self.initial_feat_dim = params['num_features'][0]

		# The output of this layer will later be reshaped to [batch, num_feature of first deconv, initial_len]
		self.input_layer = nn.Linear(hdim, self.initial_feat_dim * self.initial_len)

		conv_transpose_blocks = []

		current_len = self.initial_len
		for input_, output_, kernel_, stride_ in zip(num_features[:-1], num_features[1:], kernel, stride):
			block = []
			current_len = (current_len - 1) * stride_ + kernel_
			block.append(nn.ConvTranspose1d(input_, output_, kernel_, stride_, bias=False if params['batch_norm'] else True))
			if params['batch_norm']:
				block.append(nn.BatchNorm1d(output_))
			if params['activation'] != 'linear':
				block.append(self._activation(params['activation']))
			if params['dropout'] is not None and params['dropout'] != 0.0:
				block.append(nn.Dropout(params['dropout']))
			conv_transpose_blocks.append(nn.Sequential(*block))

		# Last layer needs kernel and stride calculated beforehand
		kernel_, stride_ = self.get_kernel_stride(current_len, self.max_len)
		block = [nn.ConvTranspose1d(num_features[-1], params['embedding_dim'], kernel_, stride_)]
		if params['batch_norm']:
			block.append(nn.BatchNorm1d(params['embedding_dim']))
		if params['activation'] != 'linear':
			block.append(self._activation(params['activation']))
		if params['dropout'] is not None and params['dropout'] != 0.0:
			block.append(nn.Dropout(params['dropout']))

		conv_transpose_blocks.append(nn.Sequential(*block))
		self.conv_transpose_blocks = nn.Sequential(*conv_transpose_blocks)

		self.use_embedding_matrix = use_embedding_matrix
		if not use_embedding_matrix:
			self.output_layer = nn.Linear(params['embedding_dim'], num_seq_labels)

	def forward(self, x, gt_input):
		x = self.input_layer(x)
		x = x.reshape(-1, self.initial_feat_dim, self.initial_len)  # "unflatten" to create [batch, feature, seq_len]
		x = self.conv_transpose_blocks(x)

		x = x.permute(0, 2, 1)  # shape=[batch, seq_len, feature]
		x = x[:, :self.max_len]  # only take the max_len sequence, as the deconvolution can only create certain discreet lengths
		if not self.use_embedding_matrix:
			x = self.output_layer(x)
		return x

	def get_kernel_stride(self, current_len, max_len):
		"""
		Determines the kernel and stride to have an output length that is higher than max_len
		https://github.com/sidhomj/DeepTCR/blob/db2c91a0422ef2c3281a594b3a7f1ac1009e3518/DeepTCR/functions/Layers.py#L240
		:params current_len:
		:params max_len:
		:return: (int, int)
		"""
		for kernel_ in list(range(4, 100)):
			for stride_ in list(range(2, 100)):
				len_after_conv_transpose = (current_len - 1) * stride_ + kernel_
				if len_after_conv_transpose >= max_len:
					return kernel_, stride_

		return AssertionError('Sequence is too long')

	def _activation(self, name):
		if name == 'relu':
			return nn.ReLU(inplace=True)
		elif name == 'leakyrelu':
			return nn.LeakyReLU(0.2, inplace=True)
		elif name == 'sigmoid':
			return nn.Sigmoid()
		elif name == 'softmax':
			return nn.Softmax()
		else:
			raise NotImplementedError(f'activation function {name} is not implemented.')


if __name__ == '__main__':
	import yaml
	with open('config/cnn_paper_oriented.yaml') as file:
		params = yaml.load(file)
	params['seq_model_hyperparams']['max_tcr_length'] = 40

	encoder = CNNEncoder(params['seq_model_hyperparams'], hdim=params['hdim'], num_seq_labels=25, use_output_layer=True).cuda()
	input_ = torch.randint(25, (10, params['seq_model_hyperparams']['max_tcr_length'])).cuda()
	x = encoder(input_, 0)

	decoder = CNNDecoder(params['seq_model_hyperparams'], hdim=params['hdim'], num_seq_labels=25).cuda()
	output = decoder(x, 0)

	print(encoder)
	print(decoder)
	print(x.shape)
	print(output.shape)



import optuna
import os

import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset
import pytorch_lightning as pl

import tcr_embedding.utils_training as utils_train

from tcr_embedding.config_optuna.mlp_count import suggest_params
from tcr_embedding.models.architectures.mlp_count_prediction import build_mlp
from tcr_embedding.models.losses.msle import MSLE


class DecisionHead(pl.LightningModule,):
    def __init__(self, params_model, n_in, n_out):
        super().__init__()
        self.save_hyperparameters()
        self.params_model = params_model
        self.network = build_mlp(params_model['mlp'], n_in, n_out).float()

        self.loss_function = MSLE()

    def forward(self, x):
        prediction = self.network(x)
        return prediction

    def training_step(self, batch, _):
        x, y_truth = batch
        x = x.float()
        y_truth = y_truth.float()
        y_pred = self.network(x)
        loss = self.loss_function(y_pred, y_truth)
        self.log('train_loss', loss)
        return loss

    def validation_step(self, batch, _):
        x, y_truth = batch
        x = x.float()
        y_truth = y_truth.float()
        y_pred = self.network(x)
        loss = self.loss_function(y_pred, y_truth)
        self.log('val_loss', loss)
        # todo r2?
        return loss

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.params_model['learning_rate'])
        return optimizer


class CustomDataset(Dataset):
    def __init__(self, x, y):
        self.x = x.astype(np.float64)
        self.y = y.astype(np.float64)

    def __len__(self):
        return len(self.x)

    def __getitem__(self, idx):
        return [self.x[idx], self.y[idx]]


def embed_data(adata, path_model, key_counts):
    model = utils_train.load_model(adata, path_model)
    embedding = model.get_latent(adata, metadata=['set'], return_mean=True)
    embedding.obsm[key_counts] = adata.obsm[key_counts]
    return embedding


def get_training_data(embedding, key_counts):
    mask_train = embedding.obs['set'] == 'train'
    x_train = embedding.X[mask_train]
    y_train = embedding.obsm[key_counts][mask_train].toarray()

    mask_val = embedding.obs['set'] == 'val'
    x_val = embedding.X[mask_val]
    y_val = embedding.obsm[key_counts][mask_val].toarray()
    return CustomDataset(x_train, y_train), CustomDataset(x_val, y_val)


def objective(trial, data, params_experiment):
    path_save = os.path.join(params_experiment['save_path'], f'trial_{trial.number}')
    params_model = suggest_params(trial)

    train_data, val_data = get_training_data(data, params_experiment['key_prediction'])

    n_in = data.X.shape[1]
    n_out = data.obsm[params_experiment['key_prediction']].shape[1]

    decision_head = DecisionHead(params_model, n_in, n_out)
    train_loader = DataLoader(train_data, batch_size=params_model['batch_size'], shuffle=True)
    val_loader = DataLoader(val_data, batch_size=params_model['batch_size'], shuffle=True)

    checkpoint_callback = pl.callbacks.ModelCheckpoint(
        dirpath=path_save,
        filename='best_model',
        save_top_k=1,
        verbose=False,
        monitor='val_loss',
        mode='min'
    )
    early_stop_callback = pl.callbacks.EarlyStopping('val_loss', patience=10)

    trainer = pl.Trainer(callbacks=[checkpoint_callback, early_stop_callback], default_root_dir=path_save)
    trainer.fit(decision_head, train_loader, val_loader)
    return trainer.checkpoint_callback.best_model_score


def run_model_selection(adata, params_experiment, num_samples, timeout=None, n_jobs=1):
    sampler = optuna.samplers.TPESampler(seed=42)  # Make the sampler behave in a deterministic way.
    storage = f'sqlite:///{params_experiment["save_path"]}.db'
    if os.path.exists(params_experiment['save_path'] + '.db'):
        os.remove(params_experiment['save_path'] + '.db')

    study = optuna.create_study(study_name=params_experiment['study_name'], sampler=sampler, storage=storage,
                                direction='minimize', load_if_exists=False)

    embedded_data = embed_data(adata, params_experiment['model_path'], params_experiment['key_prediction'])

    study.optimize(lambda trial: objective(trial, embedded_data, params_experiment),
                   n_trials=num_samples, timeout=timeout, n_jobs=n_jobs)  # todo

    best_trial = study.best_trial
    print('Best trial: ')
    print(f'  trial_{best_trial.number}')
    print(f'  Value: {best_trial.value}')



"""
This trigonometric positional embedding was adapted from:
Title: Multigrate
Date: 17th March 2021
Availability: https://github.com/theislab/multigrate
"""
from torch import nn
import torch


class MLP(nn.Module):
    def __init__(self,
                 n_inputs,
                 n_outputs,
                 hiddens=None,
                 activation='relu',
                 output_activation='linear',
                 dropout=None,
                 batch_norm=True,
                 regularize_last_layer=False):
        super(MLP, self).__init__()

        # create network architecture
        layers = []
        if hiddens is None:  # no hidden layers
            layers.append(self._fc(n_inputs, n_outputs, activation=output_activation,
                                   dropout=dropout if regularize_last_layer else None,
                                   batch_norm=regularize_last_layer))
        else:
            layers.append(self._fc(n_inputs, hiddens[0], activation=activation, dropout=dropout, batch_norm=batch_norm))  # first layer
            for l in range(1, len(hiddens)):  # inner layers
                layers.append(self._fc(hiddens[l-1], hiddens[l], activation=activation, dropout=dropout, batch_norm=batch_norm))

            layers.append(self._fc(hiddens[-1], n_outputs, activation=output_activation,
                                   dropout=dropout if regularize_last_layer else None,
                                   batch_norm=regularize_last_layer))

        self.network = nn.Sequential(*layers)

    def _fc(self, n_inputs, n_outputs, activation='leakyrelu', dropout=None, batch_norm=True):
        layers = [nn.Linear(n_inputs, n_outputs, bias=not batch_norm)]
        if batch_norm:
            layers.append(nn.BatchNorm1d(n_outputs))
        if activation != 'linear':
            layers.append(self._activation(activation))
        if dropout is not None:
            layers.append(nn.Dropout(dropout))
        return nn.Sequential(*layers)

    def _activation(self, name='leakyrelu'):
        if name == 'relu':
            return nn.ReLU(inplace=True)
        elif name == 'leakyrelu':
            return nn.LeakyReLU(0.2, inplace=True)
        elif name == 'sigmoid':
            return nn.Sigmoid()
        elif name == 'softmax':
            return nn.Softmax()
        elif name == 'exponential':
            return Exponential()
        else:
            raise NotImplementedError(f'activation function {name} is not implemented.')

    def forward(self, x):
        return self.network(x)

    def through(self, x):
        outputs = []
        for layer in self.network:
            x = layer(x)
            outputs.append(x)
        return outputs


class Exponential(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        x = torch.exp(x)
        return x


# Defines the hyper parameter of the count predictor

def suggest_params(trial):
    n_units = trial.suggest_int('rna_hidden', 100, 1000, step=100)
    num_layers = trial.suggest_int('num_layers', 1, 5, step=1)

    params = {
        'batch_size': trial.suggest_categorical('batch_size', [64, 128, 256, 512]),
        'learning_rate': trial.suggest_float('lr', 1e-5, 1e-1, log=True),

        'mlp': {
            'activation': 'leakyrelu',
            'dropout': trial.suggest_float('dropout', 0, 0.3, step=0.05),
            'batch_norm': True,
            'hidden_layers': [n_units] * num_layers
        },
    }
    return params



import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader, WeightedRandomSampler
import random

from tcr_embedding.dataloader.Dataset import JointDataset


def create_datasets(adata, val_split, metadata=None, conditional=None, labels=None, beta_only=False):
    """
    Create torch Dataset, see above for the input
    :param adata: list of adatas
    :param val_split:
    :param metadata:
    :param conditional:
    :param labels:
    :return: train_dataset, val_dataset, train_masks (for continuing training)
    """
    if metadata is None:
        metadata = []

    # Splits everything into train and val
    if val_split is not None:
        train_mask = (adata.obs[val_split] == 'train').values
    else:
        train_mask = np.ones(shape=(len(adata), ), dtype=bool)

    # Save dataset splits
    rna_train = adata.X[train_mask]
    rna_val = adata.X[~train_mask]

    if beta_only:
        tcr_seq = np.concatenate([adata.obsm['beta_seq']], axis=1)
        tcr_length = np.vstack([adata.obs['beta_len']]).T
    else:
        tcr_seq = np.concatenate([adata.obsm['alpha_seq'], adata.obsm['beta_seq']], axis=1)
        tcr_length = np.vstack([adata.obs['alpha_len'], adata.obs['beta_len']]).T
    tcr_train = tcr_seq[train_mask]
    tcr_val = tcr_seq[~train_mask]

    tcr_length_train = tcr_length[train_mask].tolist()
    tcr_length_val = tcr_length[~train_mask].tolist()

    metadata_train = adata.obs[metadata][train_mask].to_numpy()
    metadata_val = adata.obs[metadata][~train_mask].to_numpy()

    if conditional is not None:
        conditional_train = adata.obsm[conditional][train_mask]
        conditional_val = adata.obsm[conditional][~train_mask]
    else:
        conditional_train = None
        conditional_val = None

    train_dataset = JointDataset(rna_train, tcr_train, tcr_length_train, metadata_train,
                                 None, conditional_train)
    val_dataset = JointDataset(rna_val, tcr_val, tcr_length_val, metadata_val,
                               None, conditional_val)

    return train_dataset, val_dataset, train_mask


def seed_worker(worker_id):
    worker_seed = torch.initial_seed() % 2 ** 32
    np.random.seed(worker_seed)
    random.seed(worker_seed)


# <- functions for the main data loader ->
def initialize_data_loader(adata, metadata, conditional, label_key, balanced_sampling, batch_size, beta_only=False):
    train_datasets, val_datasets, train_mask = create_datasets(adata, 'set', metadata, conditional, label_key,
                                                               beta_only=beta_only)

    if balanced_sampling is None:
        train_loader = DataLoader(train_datasets, batch_size=batch_size, shuffle=True, worker_init_fn=seed_worker)
    else:
        sampling_weights = calculate_sampling_weights(adata, train_mask, class_column=balanced_sampling)
        sampler = WeightedRandomSampler(weights=sampling_weights, num_samples=len(sampling_weights),
                                        replacement=True)
        # shuffle is mutually exclusive to sampler, but sampler is anyway shuffled
        train_loader = DataLoader(train_datasets, batch_size=batch_size, shuffle=False,
                                  sampler=sampler, worker_init_fn=seed_worker)
    val_loader = DataLoader(val_datasets, batch_size=batch_size, shuffle=False)
    return train_loader, val_loader


def calculate_sampling_weights(adata, train_mask, class_column):
    """
    Calculate sampling weights for more balanced sampling in case of imbalanced classes,
    :params class_column: str, key for class to be balanced
    :params log_divisor: divide the label counts by this factor before taking the log, higher number makes the sampling more uniformly balanced
    :return: list of weights
    """
    label_counts = []

    label_count = adata[train_mask].obs[class_column].map(adata[train_mask].obs[class_column].value_counts())
    label_counts.append(label_count)

    label_counts = pd.concat(label_counts, ignore_index=True)
    label_counts = np.log(label_counts / 10 + 1)
    label_counts = 1 / label_counts

    sampling_weights = label_counts / sum(label_counts)
    return sampling_weights


# <- data loader for prediction ->
def initialize_prediction_loader(adata, metadata, batch_size, beta_only=False, conditional=None):
    prediction_dataset, _, _ = create_datasets(adata, val_split=None, conditional=conditional,
                                               metadata=metadata, beta_only=beta_only)
    prediction_loader = DataLoader(prediction_dataset, batch_size=batch_size, shuffle=False)
    return prediction_loader


# <- data loader for calculating the transcriptome from the latent space ->
def initialize_latent_loader(adata_latent, batch_size, conditional):
    if conditional is None:
        dataset = torch.utils.data.TensorDataset(torch.from_numpy(adata_latent.X))
    else:
        dataset = torch.utils.data.TensorDataset(torch.from_numpy(adata_latent.X),
                                                 torch.from_numpy(adata_latent.obsm[conditional]))
    latent_loader = torch.utils.data.DataLoader(dataset, batch_size=batch_size)
    return latent_loader



import optuna
import os
import sys
import importlib

import tcr_embedding.utils_training as utils


def get_parameter_functions(model_name, optimization_mode):
    model_name = model_name.lower()

    if optimization_mode not in ['pseudo_metric', 'knn_prediction', 'modulation_prediction']:
        model_name += '_equal'

    path_module = os.path.join(os.path.dirname(__file__), '..', 'config_optuna')
    sys.path.append(path_module)
    model_name = os.path.join(model_name)
    config_module = importlib.import_module(model_name)
    # init_params = getattr(config_module, 'init_params')
    suggest_params = getattr(config_module, 'suggest_params')
    # return init_params, suggest_params
    return suggest_params


def get_direction(optimization_mode):
    modes = {
        'pseudo_metric': 'maximize',
        'knn_prediction': 'maximize',
        'modulation_prediction': 'maximize',
        'reconstruction': 'minimize',
    }
    if optimization_mode in modes:
        return modes[optimization_mode]
    # reconstruction as default
    return 'minimize'


def complete_params_experiment(params):
    default_values = {
        'device': None,
        'balanced_sampling': None,
        'metadata': [],
        'conditional': None,
        'label_key': None,
        'n_epochs': 200,
        'kl_annealing_epochs': None,
        'early_stop': 5,
        'save_path': '../saved_models/',
        'model_name': 'moe',
    }
    for key, value in default_values.items():
        if key not in params:
            params[key] = value
    return params


def fail_save(func):
    # Keeps the HPO running, even when a training run fails.
    def wrapper(trial, adata, suggest_params, params_experiment, optimization_mode_params):
        direction = get_direction(optimization_mode_params['name'])
        try:
            return func(trial, adata, suggest_params, params_experiment, optimization_mode_params)
        except KeyboardInterrupt:
            raise
        except Exception as e:
            print(e)
            if direction == 'maximize':
                return 0.
            else:
                return 5.
    return wrapper


#@fail_save
def objective(trial, adata_tmp, suggest_params, params_experiment_base, optimization_mode_params):
    adata = adata_tmp.copy()
    params_experiment = params_experiment_base.copy()
    params_experiment = complete_params_experiment(params_experiment)
    params_experiment['save_path'] = os.path.join(params_experiment['save_path'], f'trial_{trial.number}')

    params_architecture = suggest_params(trial)

    if 'beta_only' in params_experiment and params_experiment['beta_only']:
        params_architecture['tcr']['beta_only'] = True

    if 'rna_weight' in optimization_mode_params:
        rna_weight = optimization_mode_params['rna_weight']
        params_architecture['loss_weights'] = params_architecture['loss_weights'].append(rna_weight)

    if 'use_embedding_for_cond' in params_experiment:
        params_architecture['joint']['use_embedding_for_cond'] = params_experiment['use_embedding_for_cond']
    # if 'cond_input' in params_experiment:
    #     params_architecture['joint']['cond_input'] = params_experiment['cond_input']

    comet = utils.initialize_comet(params_architecture, params_experiment)

    model = utils.select_model_by_name(params_experiment['model_name'])
    model = model(adata, params_architecture, params_experiment['balanced_sampling'], params_experiment['metadata'],
                  params_experiment['conditional'], optimization_mode_params,
                  params_experiment['label_key'], params_experiment['device'])

    model.train(params_experiment['n_epochs'], params_architecture['batch_size'], params_architecture['learning_rate'],
                params_architecture['loss_weights'], params_experiment['kl_annealing_epochs'],
                params_experiment['early_stop'], params_experiment['save_path'], comet)

    # plot UMAPs
    if comet is not None:
        model_names = ['reconstruction']
        if optimization_mode_params['name'] != 'reconstruction':
            model_names.append('metric')
        for state in model_names:
            model.load(os.path.join(params_experiment['save_path'], f'best_model_by_{state}.pt'))
            for subset in ['train', 'val']:
                adata_tmp = adata[adata.obs['set'] == subset]
                latent = model.get_latent(adata_tmp, params_experiment['metadata'], True)
                title = f'{state}_{subset}'
                figs = utils.plot_umap_list(latent, title, params_experiment['metadata'])
                if comet is not None:
                    for fig, group in zip(figs, params_experiment['metadata']):
                        comet.log_figure(f'{title}_{group}', fig)
        comet.end()
    return model.best_optimization_metric


def run_model_selection(adata, params_experiment, params_optimization, num_samples, timeout=None, n_jobs=1):
    sampler = optuna.samplers.TPESampler(seed=42)  # Make the sampler behave in a deterministic way.

    direction = get_direction(params_optimization['name'])

    storage = f'sqlite:///{params_experiment["save_path"]}.db'
    if os.path.exists(params_experiment['save_path'] + '.db'):
        os.remove(params_experiment['save_path'] + '.db')
    os.makedirs(os.path.dirname(params_experiment['save_path']), exist_ok=True)

    study = optuna.create_study(study_name=params_experiment['study_name'], sampler=sampler, storage=storage,
                                direction=direction, load_if_exists=False)

    suggest_params = get_parameter_functions(params_experiment['model_name'], params_optimization['name'])
    # study.enqueue_trial(init_params)
    study.optimize(lambda trial: objective(trial, adata, suggest_params, params_experiment, params_optimization),
                   n_trials=num_samples, timeout=timeout, n_jobs=n_jobs)

    pruned_trials = [t for t in study.trials if t.state == optuna.structs.TrialState.PRUNED]
    complete_trials = [t for t in study.trials if t.state == optuna.structs.TrialState.COMPLETE]

    print('Study statistics:')
    print(f'  Number of finished trials: {len(study.trials)}')
    print(f'  Number of pruned trials: {len(pruned_trials)}')
    print(f'  Number of complete trials: {len(complete_trials)}')

    best_trial = study.best_trial
    print('Best trial: ')
    print(f'  trial_{best_trial.number}')
    print(f'  Value: {best_trial.value}')



import numpy as np
import random
import tcr_embedding.evaluation.WrapperFunctions as Wrapper
import tcr_embedding.evaluation.Metrics as Metrics

import scanpy as sc


def run_imputation_evaluation(data_full, embedding_function, query_source='val',
                              num_neighbors=5, label_pred='binding_name'):
    """
    Function for evaluating the embedding quality based upon imputation in the 10x dataset
    :param data_full: anndata object containing the full cell data (TCR + Genes) (train, val, test)
    :param embedding_function: function calculating the latent space for a single input
    :param query_source: str 'val' or 'test' to choose between evaluation mode
    :param num_neighbors: amount of neighbors for knn classification
    :param label_pred: label of the collumn used for prediction
    :return: dictionary {metric: summary} containing the evaluation scores
    """
    data_atlas = data_full[data_full.obs['set'] == 'train']
    data_query = data_full[data_full.obs['set'] == query_source]

    assert len(data_query) > 0, 'Empty query set. Specifier are "val" or "test"'

    embedding_atlas = embedding_function(data_atlas)
    embedding_query = embedding_function(data_query)

    scores = get_imputation_scores(embedding_atlas, embedding_query,
                                   data_atlas.obs[label_pred], data_query.obs[label_pred],
                                   num_neighbors=num_neighbors)
    return scores


def get_imputation_scores(embedding_atlas, embedding_query, label_atlas, label_embedding, num_neighbors=5):
    """
    Calculate evaluation scores based on imputation, so far knn classification report and Recall@k
    :param embedding_atlas: numpy array (num cells, hidden dim) of the atlas dataset
    :param embedding_query: see above of the query dataset
    :param label_atlas: epitope specificity by cell of the atlas
    :param label_embedding: epitope specificity by cell of the query set
    :param num_neighbors: amount of neighbors used for knn classification
    :return: dictionary {metric_name: metric_score}
    """
    summary = {}

    # Recall at k
    ks = [1, 10, 100]
    # recalls = Metrics.get_recall_at_k(embedding_atlas, embedding_query, label_atlas, label_embedding, ks)
    # summary['R@k'] = recalls

    # kNN score
    knn_score = Metrics.get_knn_classification(embedding_atlas, embedding_query, label_atlas, label_embedding,
                                           num_neighbors=num_neighbors, weights='distance')
    summary['knn'] = knn_score
    return summary


if __name__ == '__main__':
    """ For testing purposes """
    test_embedding_func = Wrapper.get_random_prediction_function(hidden_dim=800)
    print('Reading data')
    test_data = sc.read('../../data/10x_CD8TC/v5_train_val_test.h5ad')
    random.seed(29031995)
    # test_data = test_data[[random.randint(0, test_data.shape[0]-1) for _ in range(int(test_data.shape[0]*0.1))]]
    print('Start evaluation')
    res = run_imputation_evaluation(test_data, test_embedding_func, query_source='val',
                                    use_non_binder=True, use_reduced_binders=True)
    print(res)



"""
This code was partially adapted from:
Date: 15th March 2021
Availability: https://github.com/theislab/multigrate
"""
import torch


class KLD(torch.nn.Module):
    def __init__(self, reduction='mean'):
        super(KLD, self).__init__()
        self.reduction = reduction

    def forward(self, mu, logvar, mu_2=None, logvar_2=None):
        """
        Calculate the Kullbach-Leibler-Divergence between two Gaussians
        :param mu: mean of the first Gaussian
        :param logvar: log(var) of the first Gaussian
        :param mu_2: mean of the second Gaussian (default: 0)
        :param logvar_2: log(var) of the second Gaussian (default: 1)
        :return: loss value
        """
        if mu_2 is None or logvar_2 is None:
            kl = self.univariate_kl_loss(mu, logvar)
        else:
            kl = self.general_kl_loss(mu, logvar, mu_2, logvar_2)
        if self.reduction == 'mean':
            kl = torch.mean(kl)
        elif self.reduction == 'sum':
            kl = torch.sum(kl)
        else:
            raise NotImplementedError(f'reduction method {self.reduction} is not implemented.')
        return kl

    def univariate_kl_loss(self, mu, logvar):
        """
        KL loss between the input and a 0 mean, uni-variance Gaussian
        :param mu: mean of the distribution
        :param logvar: log variance of the distribution
        :return: Kulbach Leibler divergence between distribution and Gaussian
        """
        kl = 0.5 * torch.sum(torch.exp(logvar) + mu**2 - 1. - logvar, dim=1)
        return kl

    def general_kl_loss(self, mu_1, logvar_1, mu_2, logvar_2):
        """
        KL loss between two distributions
        :param mu_1: mean of the first distribution
        :param logvar_1: log variance of the first distribution
        :param mu_2: mean of the second distribution
        :param logvar_2: log variane of the second distribution
        :return: Kulbach Leibler divergence loss between the two distributions
        """
        kl = logvar_2 - logvar_1 + torch.exp(logvar_1)/torch.exp(logvar_2) + (mu_1-mu_2)**2/torch.exp(logvar_2)-1
        kl = 0.5 * torch.sum(kl)
        return kl



"""
File containing the metrics for evaluating the embedding space
"""
import numpy as np
import torch

from sklearn.neighbors import KNeighborsClassifier

from sklearn.metrics import classification_report, silhouette_score, adjusted_mutual_info_score
from sklearn.metrics import normalized_mutual_info_score, adjusted_rand_score

from scipy import stats


def get_knn_classification(data_atlas, data_query, labels_atlas, labels_query, num_neighbors=5, weights='distance'):
    """
    Evaluates with kNN based on scikit-learn
    :param data_atlas: numpy array (num_cells, hidden_size) embeddings of the atlas data
    :param data_query: numpy array (num_cells, hidden_size) embeddings of the query data
    :param labels_atlas: list (num_cells) labels of the atlas data
    :param labels_query: list (num_cells) labels of the query data
    :param num_neighbors: amount of neighbors used for kNN
    :param weights: kNN weighting,
    :return:
    """

    clf = KNeighborsClassifier(num_neighbors, weights)
    clf.fit(data_atlas, labels_atlas)

    labels_predicted = clf.predict(data_query)
    report = classification_report(labels_query, labels_predicted, output_dict=True)
    return report


def get_silhouette_scores(embeddings, labels_predicted):
    """
    Calculates the Silhouette score as internal cluster evaluation
    :param embeddings: numpy array (n_samples, dim_hidden) containing the latent embedding
    :param labels_predicted: predicted labels based on clustering in latent space
    :return:
    """
    try:
        score = silhouette_score(embeddings, labels_predicted, metric='euclidean', random_state=29031995)
    # If the number of cluster is 1, then ASW can't be calculated and raises an Error
    except:
        score = -99
    return score


def get_adjusted_mutual_information(labels_true, labels_predicted):
    """
    Calculates the AMI score as external cluster evaluation
    :param labels_true: ground truth labels for external cluster evaluation
    :param labels_predicted: predicted labels based on clustering in latent space
    :return: Adjusted mutual information score
    """
    scores = adjusted_mutual_info_score(labels_true, labels_predicted)
    return scores


def get_normalized_mutual_information(labels_true, labels_predicted):
    """
    Calculates the NMI score as external cluster evaluation
    :param labels_true: ground truth labels for external cluster evaluation
    :param labels_predicted: predicted labels based on clustering in latent space
    :return: Normalized mutual information score
    """
    scores = normalized_mutual_info_score(labels_true, labels_predicted, average_method='arithmetic')
    return scores


def get_adjusted_random_score(labels_true, labels_predicted):
    """
    Calculates the AMI score as external cluster evaluation
    :param labels_true: ground truth labels for external cluster evaluation
    :param labels_predicted: predicted labels based on clustering in latent space
    :return: Adjusted mutual information score
    """
    scores = adjusted_rand_score(labels_true, labels_predicted)
    return scores


def get_square_pearson(ground_truth, prediction):
    try:
        x = np.average(ground_truth.X.A, axis=0)
    except AttributeError:
        x = np.average(ground_truth.X, axis=0)
    try:
        y = np.average(prediction.X.A, axis=0)
    except AttributeError:
        y = np.average(prediction.X, axis=0)
    m, b, r_value, p_value, std_err = stats.linregress(x, y)
    r_squared = r_value**2
    report = {
        'm': m,
        'b': b,
        'r_value': r_value,
        'p_value': p_value,
        'std_err': std_err,
        'r_squared': r_squared
    }
    return report


def get_knn_f1_within_set(latent, column_name):
    con = latent.obsp['connectivities'].A.astype(np.bool)
    nearest_neighbor_label = [latent.obs[column_name].values[row].tolist()[0] for row in con]
    labels_true = latent.obs[column_name].values
    if torch.is_tensor(nearest_neighbor_label[0]):
        nearest_neighbor_label = [str(el.item()) for el in nearest_neighbor_label]
        labels_true = [str(el.item()) for el in labels_true]
    result = classification_report(labels_true, nearest_neighbor_label, output_dict=True)
    result = result['weighted avg']['f1-score']
    return result



from tcr_embedding.models.architectures.mlp import MLP


def build_mlp(params, n_in, n_out):
    print(params)
    mlp = MLP(n_inputs=n_in,
              n_outputs=n_out,
              hiddens=params['hidden_layers'],
              activation=params['activation'],
              output_activation='exponential',
              dropout=params['dropout'],
              batch_norm=params['batch_norm'],
              regularize_last_layer=False)
    return mlp



from tcr_embedding.evaluation.PertubationPrediction import evaluate_pertubation
import numpy as np
import scanpy as sc
import time

sc.settings.verbosity = 0


def predict_pertubation(latent_train, latent_val, model, column_perturbation, indicator_perturbation, var_names,
                        return_latent=False, per_type=False, col_type=None):
    """
    Predict the effect of pertubation on transcriptome level
    :param latent_train: adata object containing the latent spaces of the training dataset
    :param latent_val: adata object containing the latent spaces of the validaiton dataset
    :param model: model to predict transcritome from latent space
    :param column_perturbation: str, column in the adata objects indicating the pertubation
    :param indicator_perturbation: str, value for the 'pre' state of the perturbation
    :param var_names: list containing the gene names
    :return: adata object, containing the predicted transcriptome profile after perturbation
    """
    # todo delta per cell type? # upsampling?
    if not per_type:
        delta = get_delta(latent_train, column_perturbation, indicator_perturbation)
        if col_type is None:
            pass
        else:
            deltas = []

            cats_pre = latent_train[latent_train.obs[column_perturbation] == indicator_perturbation].obs[
                col_type].value_counts()
            cats_pre = [cat for cat, count in cats_pre.items() if count > 0]
            cats_post = latent_train[latent_train.obs[column_perturbation] != indicator_perturbation].obs[
                col_type].value_counts()
            cats_post = [cat for cat, count in cats_post.items() if count > 0]
            cats_both = [cat for cat in cats_pre if cat in cats_post]

            for t in cats_both:
                d = get_delta(latent_train[latent_train.obs[col_type]==t], column_perturbation, indicator_perturbation)
                deltas.append(d)
            delta = np.vstack(deltas).mean(axis=0)
        adata_pred = sc.AnnData(latent_val.X + delta, obs=latent_val.obs.copy())
    else:
        adatas_pred = []
        for t in latent_train.obs[per_type].unique():
            delta = get_delta(latent_train[latent_train.obs[per_type] == t],
                              column_perturbation, indicator_perturbation)
            adata_tmp = latent_val[latent_val.obs[per_type] == t].copy()
            adata_new = sc.AnnData(adata_tmp.X + delta, obs=adata_tmp.obs)
            adatas_pred.append(adata_new)
        adata_pred = sc.concat(adatas_pred)
    if return_latent:
        return adata_pred
    adata_pred = model.predict_rna_from_latent(adata_pred, metadata=adata_pred.obs.columns)
    adata_pred.var_names = var_names
    return adata_pred


def get_delta(adata_latent, column_perturbation, indicator_perturbation):
    """
    Calculate the difference vector between pre and post perturbation in the latent space
    :param adata_latent: adata oject, containing the latent space representation of the training data
    :param column_perturbation: str, column in the adata object indicating the pertubation
    :param indicator_perturbation: str, value for the 'pre' state of the profile after perturbation
    :return: numpy array, difference vectore
    """
    mask_pre = adata_latent.obs[column_perturbation] == indicator_perturbation
    latent_pre = adata_latent[mask_pre]
    latent_post = adata_latent[~mask_pre]
    avg_pre = np.mean(latent_pre.X, axis=0)
    avg_post = np.mean(latent_post.X, axis=0)
    delta = avg_post - avg_pre
    return delta


def run_scgen_cross_validation(adata, column_fold, model, column_perturbation, indicator_perturbation):
    """
    Runs perturbation prediction over a specified fold column and evaluates the results
    :param adata: adata object, of the raw data
    :param column_fold: str, indicating the column over which to cross validate
    :param model: model used for latent space generation
    :param column_perturbation: str, column in the adata object indicating the pertubation
    :param indicator_perturbation: str, value for the 'pre' state of the profile after perturbation
    :return: dict, summary over performance on the different splits and aggregation
    """
    latent_full = model.get_latent(adata, metadata=[column_fold, column_perturbation])

    summary_performance = {}
    rs_squared = []

    cats_pre = adata[adata.obs[column_perturbation] == indicator_perturbation].obs[column_fold].value_counts()
    cats_pre = [cat for cat, count in cats_pre.items() if count > 10]
    cats_post = adata[adata.obs[column_perturbation] != indicator_perturbation].obs[column_fold].value_counts()
    cats_post = [cat for cat, count in cats_post.items() if count > 10]
    cats_both = [cat for cat in cats_pre if cat in cats_post]

    for fold in cats_both:
        mask_train = latent_full.obs[column_fold] != fold
        latent_train = latent_full[mask_train]
        latent_val = latent_full[~mask_train]

        latent_val_pre = latent_val[latent_val.obs[column_perturbation] == indicator_perturbation]
        pred_val_post = predict_pertubation(latent_train, latent_val_pre, model,
                                            column_perturbation, indicator_perturbation,
                                            var_names=adata.var_names, col_type=column_fold)

        score = evaluate_pertubation(adata[adata.obs[column_fold] == fold].copy(), pred_val_post, None,
                                     column_perturbation, indicator=indicator_perturbation)

        for key, value in score.items():
            summary_performance[f'{fold}_key'] = value
        rs_squared.append(score['all_genes']['r_squared'])

    summary_performance['avg_r_squared'] = sum(rs_squared) / len(rs_squared)
    return summary_performance



try:
    from comet_ml import Experiment
except:
    pass
import scanpy as sc
import os
import random
import torch
import numpy as np

from tcr_embedding.models.mixture_modules.rna_model import RnaModel
from tcr_embedding.models.mixture_modules.separate_model import SeparateModel
from tcr_embedding.models.mixture_modules.poe import PoEModel
from tcr_embedding.models.mixture_modules.moe import MoEModel


def fix_seeds(random_seed=42):
    torch.manual_seed(random_seed)
    np.random.seed(random_seed)
    random.seed(random_seed)


def load_data(source='10x'):
    """
    Loads the whole dataset from a defined source.
    :param source: str indicting the dataset or filename starting from the data_folder
    :return: adata object
    """
    path_current = os.path.dirname(__file__)
    path_base = os.path.join(path_current, '../data/')

    source = source.lower()
    if source == '10x':
        path_source = '10x_CD8TC/v7_avidity.h5ad'
    elif source == 'bcc':
        path_source = 'BCC/06_bcc_highly_var_5000.h5ad'
    elif source == 'scc':
        path_source = 'SCC/06_scc_highly_var_5000.h5ad'
    elif source == 'covid':
        path_source = 'Covid/04_covid_highly_var_5000.h5ad'
    elif source == 'haniffa':
        path_source = 'Haniffa/v3_conditional.h5ad'
    elif source == 'haniffa_bcr':
        path_source = 'Haniffa/02_bcrs_annoated.h5ad'
    elif source == 'borcherding_test':
        path_source = 'Borcherding/04_borch_annotated_test.h5ad'
    elif source == 'borcherding':
        path_source = 'Borcherding/04_borch_annotated.h5ad'
    elif source == 'bcells_covid':
        path_source = 'Bcells_Covid/02_bcrs_annoated.h5ad'
    else:
        path_source = source
    path_file = os.path.join(path_base, path_source)

    try:
        data = sc.read_h5ad(path_file)
    except FileNotFoundError:
        raise FileNotFoundError(f'Data file not found at {path_file}. '
                                f'Please specify data source by "10x", "bcc", "scc", "haniffa", or "covid". '
                                f'Alternatively, as filename starting from the data folder.')
    return data


def load_model(adata, path_model, base_path=None):
    if base_path is None:
        base_path = os.path.dirname(__file__)
        path_model = os.path.join(base_path, '..', path_model)
    else:
        path_model = os.path.join(base_path, path_model)
    model_file = torch.load(path_model)

    params_architecture = model_file['params_architecture']
    balanced_sampling = model_file['balanced_sampling']
    metadata = model_file['metadata']
    conditional = model_file['conditional']
    optimization_mode_params = model_file['optimization_mode_params']
    label_key = model_file['label_key']

    model_class = select_model_by_name(model_file['model_type'])
    model = model_class(adata, params_architecture, balanced_sampling, metadata, conditional,
                        optimization_mode_params, label_key)
    model.load(path_model)
    return model


def initialize_comet(params_architecture, params_experiment):
    if params_experiment['comet_workspace'] is None:
        return None
    path_key = os.path.join(os.path.dirname(__file__), '../config/API_key.txt')
    with open(path_key) as f:
        comet_key = f.read()

    experiment_name = params_experiment['study_name']
    workspace = params_experiment['comet_workspace']
    experiment = Experiment(api_key=comet_key, workspace=workspace, project_name=experiment_name)

    experiment.log_parameters(params_experiment, prefix='fixed_')

    for tag in ['rna', 'tcr', 'rna', 'joint']:
        if tag in params_architecture:
            experiment.log_parameters(params_architecture[tag], prefix=tag)
    return experiment


def select_model_by_name(model_name):
    """
    Select between modeltypes (e.g. single, concat, poe, moe, ...) by an identifier
    :param model_name:  str indicating the model type, type is chosen, when following indicator are part of the string
                        'concat': for individual alpha beta chain representation
                        'moe': mixture of experts
                        'poe': product of experts
    :return: class of the corresponding model
    """
    if model_name.startswith('debug_'):
        model_name = model_name.replace('debug_', '')

    init_dict = {
        'rna': RnaModel,
        'concat': SeparateModel,
        'separate': SeparateModel,
        'tcr': SeparateModel,
        'moe': MoEModel,
        'poe': PoEModel,
    }
    model_name = model_name.lower()
    if model_name in init_dict:
        return init_dict[model_name]
    else:
        raise ValueError('Please specify a valid model name')


def determine_marker_genes(adata, resolution, visualize=False, filter_tcr=True):
    adata = adata[adata.obs['set'] == 'train']
    sc.pp.neighbors(adata, use_rep='X')
    sc.tl.leiden(adata, resolution=resolution)
    if visualize:
        sc.tl.umap(adata)  # for visualization only
        sc.pl.umap(adata, color='leiden')

    # Filter TRA, TRB and TRG which forms the T-cell receptor
    if filter_tcr:
        adata = adata[:, ~adata.var.index.str.contains('TRA')]
        adata = adata[:, ~adata.var.index.str.contains('TRB')]
        adata = adata[:, ~adata.var.index.str.contains('TRG')]

    sc.tl.rank_genes_groups(adata, groupby='leiden')
    highly_variable = list(adata.uns['rank_genes_groups']['names'][0])
    return highly_variable


def plot_umap_list(adata, title, color_groups):
    """
    Plots UMAPS based with different coloring groups
    :param adata: Adata Object containing a latent space embedding
    :param title: Figure title
    :param color_groups: Column name in adata.obs used for coloring the UMAP
    :return:
    """
    try:
        if adata.X.shape[1] == 2:
            adata.obsm['X_umap'] = adata.X
        else:
            sc.pp.neighbors(adata, use_rep='X')
            sc.tl.umap(adata)
        figures = []
        for group in color_groups:
            fig = sc.pl.umap(adata, color=group, title=title+'_'+group, return_fig=True)
            fig.tight_layout()
            figures.append(fig)
        return figures
    except ValueError as e:
        print(e)
        return []



import random
import tcr_embedding.evaluation.WrapperFunctions as Wrapper
import tcr_embedding.evaluation.Metrics as Metrics

import scanpy as sc
from anndata import AnnData


def run_clustering_evaluation(data_full, embedding_function, source_data='val', name_label='clonotype',
                              cluster_params=None, visualize=False):
    """
    Function for evaluating the embedding quality based upon imputation in the 10x dataset
    :param data_full: anndata object containing the full cell data (TCR + Genes) (train, val, test)
    :param embedding_function: function calculating the latent space for a single input
    :param source_data: list or str 'val' or 'test' to choose between evaluation mode
    :param name_label: column in data_full.obs used as ground truth for external cluster evaluation
    :param cluster_params: parameters for leiden clustering, None for default
    :return: dictionary {metric: summary} containing the evaluation scores
    """
    try:
        if type(source_data) is list:
            data_eval = data_full[data_full.obs['set'].isin(source_data)]
        elif source_data == 'all':
            data_eval = data_full
        else:
            data_eval = data_full[data_full.obs['set'] == source_data]

        assert len(data_eval) > 0, 'Empty data set. Specifier are "val" or "test"'

        embeddings = embedding_function(data_eval)
        embeddings_adata = AnnData(embeddings)
        embeddings_adata.obs[name_label] = data_eval.obs[name_label].to_numpy()

        labels_true = data_eval.obs[name_label].to_numpy()
        labels_predicted = predict_clustering(embeddings_adata, cluster_params, visualize, name_label)

        scores = get_clustering_scores(embeddings, labels_true, labels_predicted)
    except ValueError:
        # If training runs haywire we end up here during evaluation
        print('Latent space contains NaN. Set worst clustering scores')
        scores = {
            'ASW': -1,
            'ARI': -1,
            'NMI': 0
        }
    return scores


def predict_clustering(adata, params, visualize=False, name_label=None):
    """
    Calculate cluster labels based on leiden algorithm
    :param adata: annData with X containing the embeddings (num_samples, dim_hidden)
    :param params: parameter of leiden clustering
    :return: numpy array (num_cells,) containing cluster annotation
    """
    if params is None:
        params = {'resolution': 1,
                  'num_neighbors': 5}
    sc.pp.neighbors(adata, n_neighbors=params['num_neighbors'], use_rep='X', random_state=29031995)

    sc.tl.leiden(adata, resolution=params['resolution'], random_state=29031995)
    if visualize:
        sc.tl.umap(adata)
        palette = None
        sc.pl.umap(adata, color=name_label, palette=palette)
        sc.pl.umap(adata, color='leiden', title=f'resolution = {params["resolution"]}')

    return adata.obs['leiden'].to_numpy()


def get_clustering_scores(embeddings, labels_true, labels_predicted):
    """
    Calculate evaluation scores based on clustering.
    Silhouette score for internal, Adjusted Mutual Information as external evaluation.
    :param embeddings: numpy array (num cells, hidden dim) of the cell dataset
    :param labels_true: ground truth cluster annotation for external evaluation
    :param labels_predicted: predicted cluster annotation for internal / external evaluation
    :return: dictionary {metric_name: metric_score}
    """
    summary = {
        'ASW': Metrics.get_silhouette_scores(embeddings, labels_predicted),
        'ARI': Metrics.get_adjusted_random_score(labels_true, labels_predicted),
        'NMI': Metrics.get_normalized_mutual_information(labels_true, labels_predicted)
    }
    return summary


if __name__ == '__main__':
    """ For testing purposes """

    test_embedding_func = Wrapper.get_random_prediction_function(hidden_dim=800)
    print('Reading data')
    test_data = sc.read('../../data/10x_CD8TC/v5_train_val_test.h5ad')
    random.seed(29031995)
    # test_data = test_data[[random.randint(0, test_data.shape[0]-1) for _ in range(int(test_data.shape[0]*0.1))]]
    print('Start Cluster evaluation')
    res = run_clustering_evaluation(test_data, test_embedding_func, source_data='val', name_label='clonotype')
    print(res)



from setuptools import setup, find_packages

setup(
    name='mvtcr',
    version='0.1.1',
    packages=find_packages(include=['tcr_embedding', 'tcr_embedding.*']),
    install_requires=[
        'scanpy>=1.7.0', 
        'anndata>=0.7.5', 
        'numba>=0.52.0', 
        'pandas>=1.2.3',
        'numpy>=1.20.3', 
        'scikit-learn>=0.24.1', 
        'tqdm', 
        'umap-learn>=0.5.1', 
        'torch>=1.8.0'],
    extras_require={
        'paper': [
            'scanpy==1.7.0', 
            'scirpy==0.6.1', 
            'numba==0.52.0', 
            'pandas==1.2.3', 
            'anndata==0.7.5', 
            'leidenalg==0.8.4', 
            'comet-ml', 
            'numpy==1.20.3', 
            'scikit-learn==0.24.1', 
            'scrublet==0.2.3', 
            'tqdm', 
            'optuna', 
            'umap-learn==0.5.1', 
            'torch==1.8.0'
            ],
        'full': [
            'comet-ml', 
            'scrublet==0.2.3', 
            'optuna', 
            'scirpy==0.6.1', 
            'leidenalg>=0.8.4'
            ],
    }
)


import operator

from tcr_embedding.evaluation.WrapperFunctions import get_model_prediction_function
from tcr_embedding.evaluation.Imputation import run_imputation_evaluation


def report_knn_prediction(adata, model, optimization_mode_params, epoch, comet):
    """
    Report the objective metric of the 10x dataset for hyper parameter optimization.
    :param epoch: epoch number for logging
    :return: Reports externally to comet, saves model.
    """
    test_embedding_func = get_model_prediction_function(model)
    summary = run_imputation_evaluation(adata, test_embedding_func, query_source='val',
                                        label_pred=optimization_mode_params['prediction_column'])
    metrics = summary['knn']

    if comet is not None:
        for antigen, metric in metrics.items():
            if antigen != 'accuracy':
                comet.log_metrics(metric, prefix=antigen, epoch=epoch)
            else:
                comet.log_metric('accuracy', metric, epoch=epoch)
    return metrics['weighted avg']['f1-score'], operator.gt



import warnings

warnings.filterwarnings('ignore')  # "error", "ignore", "always", "default", "module" or "once"

import torch
import torch.nn as nn
import os
import numpy as np
import pandas as pd
from tqdm import tqdm
from collections import defaultdict
import scanpy as sc
from abc import ABC, abstractmethod

from .losses.kld import KLD

from tcr_embedding.dataloader.DataLoader import initialize_data_loader, initialize_latent_loader
from tcr_embedding.dataloader.DataLoader import initialize_prediction_loader

from .optimization.knn_prediction import report_knn_prediction
from .optimization.modulation_prediction import report_modulation_prediction
from .optimization.pseudo_metric import report_pseudo_metric


class VAEBaseModel(ABC):
	def __init__(self,
				 adata,
				 params_architecture,
				 balanced_sampling='clonotype',
				 metadata=None,
				 conditional=None,
				 optimization_mode_params=None,
				 label_key=None,
				 device=None):
		"""
		VAE Base Model, used for both single and joint models
		:param adata: list of adatas containing train and val set
		:param conditional: str or None, if None a normal VAE is used, if str then the str determines the adata.obsm[conditional] as conditioning variable
		:param metadata: list of str, list of metadata that is needed, not really useful at the moment
		:param balanced_sampling: None or str, indicate adata.obs column to balance
		:param optimization_mode_params: dict carrying the mode specific parameters
		"""
		self.adata = adata
		self.params_architecture = params_architecture
		self.balanced_sampling = balanced_sampling
		self.metadata = metadata
		self.conditional = conditional
		self.optimization_mode_params = optimization_mode_params
		self.label_key = label_key
		self.device = device

		self.params_tcr = None
		self.params_rna = None
		self.params_joint = params_architecture['joint']
		self.params_supervised = None
		self.beta_only = False

		if 'tcr' in params_architecture:
			self.params_tcr = params_architecture['tcr']
			if 'beta_only' in self.params_tcr:
				self.beta_only = self.params_tcr['beta_only']
		if 'rna' in params_architecture:
			self.params_rna = params_architecture['rna']
		if 'supervised' in params_architecture:
			self.params_supervised = params_architecture['supervised']

		if self.params_tcr is None and self.params_rna is None:
			raise ValueError('Please specify either tcr, rna, or both hyperparameters.')

		self.aa_to_id = adata.uns['aa_to_id']

		if self.device is None:
			self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

		self._train_history = defaultdict(list)
		self._val_history = defaultdict(list)

		# counters
		self.best_optimization_metric = None
		self.best_loss = None
		self.no_improvements = 0

		# loss functions
		self.loss_function_rna = nn.MSELoss()
		self.loss_function_tcr = nn.CrossEntropyLoss(ignore_index=self.aa_to_id['_'])
		self.loss_function_kld = KLD()
		self.loss_function_class = nn.CrossEntropyLoss()

		# training params
		self.batch_size = params_architecture['batch_size']
		self.loss_weights = None
		self.comet = None
		self.kl_annealing_epochs = None

		# Model
		self.model_type = None
		self.model = None
		self.optimizer = None
		self.supervised_model = None
		if self.label_key is not None:
			assert self.params_supervised is not None, 'Please specify parameters for supervised model'
			self.supervised_model = None  # todo

		# datasets
		if metadata is None:
			metadata = []
		if balanced_sampling is not None and balanced_sampling not in metadata:
			metadata.append(balanced_sampling)
		self.data_train, self.data_val = initialize_data_loader(adata, metadata, conditional, label_key,
																balanced_sampling, self.batch_size,
																beta_only=self.beta_only)

	def change_adata(self, new_adata):
		self.adata = new_adata
		self.aa_to_id = new_adata.uns['aa_to_id']
		if self.balanced_sampling is not None and self.balanced_sampling not in self.metadata:
			self.metadata.append(self.balanced_sampling)

		self.data_train, self.data_val = initialize_data_loader(new_adata, self.metadata, self.conditional, self.label_key,
																self.balanced_sampling, self.batch_size,
																beta_only=self.beta_only)

	def add_new_embeddings(self, num_new_embeddings):
		cond_emb_tmp = self.model.cond_emb.weight.data
		self.model.cond_emb = torch.nn.Embedding(cond_emb_tmp.shape[0]+num_new_embeddings, cond_emb_tmp.shape[1])
		self.model.cond_emb.weight.data[:cond_emb_tmp.shape[0]] = cond_emb_tmp

	def freeze_all_weights_except_cond_embeddings(self):
		"""
		Freezes conditional embedding weights to train in scArches style, since training data doesn't include
		previous labels, those embeddings won't be updated
		"""
		for param in self.model.parameters():
			param.requires_grad = False

		self.model.cond_emb.weight.requires_grad = True

	def unfreeze_all(self):
		for param in self.model.parameters():
			param.requires_grad = True

	def train(self,
			  n_epochs=200,
			  batch_size=512,
			  learning_rate=3e-4,
			  loss_weights=None,
			  kl_annealing_epochs=None,
			  early_stop=None,
			  save_path='../saved_models/',
			  comet=None):
		"""
		Train the model for n_epochs
		:param n_epochs: None or int, number of epochs to train, if None n_iters needs to be int
		:param batch_size: int, batch_size
		:param learning_rate: float, learning rate
		:param loss_weights: list of floats, loss_weights[0]:=weight or scRNA loss, loss_weights[1]:=weight for TCR loss, loss_weights[2] := KLD Loss
		:param kl_annealing_epochs: int or None, int number of epochs until kl reaches maximum warmup, if None this value is set to 30% of n_epochs
		:param early_stop: int, stop training after this number of epochs if val loss is not improving anymore
		:param save_path: str, path to directory to save model
		:param comet: None or comet_ml.Experiment object
		:return:
		"""
		self.batch_size = batch_size
		self.loss_weights = loss_weights
		self.comet = comet
		self.kl_annealing_epochs = kl_annealing_epochs
		assert 3 <= len(loss_weights) <= 4, 'Length of loss weights need to be either 3 or 4.'

		try:
			os.makedirs(save_path)  # Create directory to prevent Error while saving model weights
		except OSError:
			pass

		if kl_annealing_epochs is None:
			self.kl_annealing_epochs = int(0.3 * n_epochs)

		# raise ValueError(f'length of loss_weights must be 3, 4 (supervised) or None.')

		self.model = self.model.to(self.device)
		self.optimizer = torch.optim.Adam(params=self.model.parameters(), lr=learning_rate)

		for epoch in tqdm(range(n_epochs)):
			self.model.train()
			train_loss_summary = self.run_epoch(epoch, phase='train')
			self.log_losses(train_loss_summary, epoch)

			self.model.eval()
			with torch.no_grad():
				val_loss_summary = self.run_epoch(epoch, phase='val')
				self.log_losses(val_loss_summary, epoch)
				self.additional_evaluation(epoch, save_path)

			if self.do_early_stopping(val_loss_summary['val Loss'], early_stop, save_path, epoch):
				break

	def run_epoch(self, epoch, phase='train'):
		if phase == 'train':
			data = self.data_train
		else:
			data = self.data_val
		loss_total = []
		rna_loss_total = []
		tcr_loss_total = []
		kld_loss_total = []
		cls_loss_total = []
		cls_acc_total = []

		for rna, tcr, seq_len, _, labels, conditional in data:
			if rna.shape[0] == 1 and phase == 'train':
				continue  # BatchNorm cannot handle batches of size 1 during training phase
			rna = rna.to(self.device)
			tcr = tcr.to(self.device)

			if self.conditional is not None:
				conditional = conditional.to(self.device)
			else:
				conditional = None

			z, mu, logvar, rna_pred, tcr_pred = self.model(rna, tcr, seq_len, conditional)
			kld_loss, z = self.calculate_kld_loss(mu, logvar, epoch)
			rna_loss, tcr_loss = self.calculate_loss(rna_pred, rna, tcr_pred, tcr)
			loss = kld_loss + rna_loss + tcr_loss

			if self.supervised_model is not None:
				prediction_label = self.forward_supervised(z)
				cls_acc = torch.sum(torch.eq(prediction_label, labels))
				cls_loss = self.calculate_classification_loss(prediction_label, labels)
				loss += cls_loss
				cls_loss_total.append(cls_loss)
				cls_acc_total.append(cls_acc)

			if phase == 'train':
				self.run_backward_pass(loss)

			loss_total.append(loss)
			rna_loss_total.append(rna_loss)
			tcr_loss_total.append(tcr_loss)
			kld_loss_total.append(kld_loss)

			if torch.isnan(loss):
				print(f'ERROR: NaN in loss.')
				return

		loss_total = torch.stack(loss_total).mean().item()
		rna_loss_total = torch.stack(rna_loss_total).mean().item()
		tcr_loss_total = torch.stack(tcr_loss_total).mean().item()
		kld_loss_total = torch.stack(kld_loss_total).mean().item()

		summary_losses = {f'{phase} Loss': loss_total,
						  f'{phase} RNA Loss': rna_loss_total,
						  f'{phase} TCR Loss': tcr_loss_total,
						  f'{phase} KLD Loss': kld_loss_total}

		if self.supervised_model is not None:
			cls_loss_total = torch.stack(cls_loss_total).mean().item()
			summary_losses[f'{phase} CLS Loss'] = cls_loss_total
			summary_losses[f'{phase} CLS Accuracy'] = cls_acc_total

		return summary_losses

	def log_losses(self, summary_losses, epoch):
		if self.comet is not None:
			self.comet.log_metrics(summary_losses, epoch=epoch)

	def run_backward_pass(self, loss):
		self.optimizer.zero_grad()
		loss.backward()
		if self.optimization_mode_params is not None and 'grad_clip' in self.optimization_mode_params:
			nn.utils.clip_grad_value_(self.model.parameters(), self.optimization_mode_params['grad_clip'])
		self.optimizer.step()

	def additional_evaluation(self, epoch, save_path):
		if self.optimization_mode_params is None:
			return
		name = self.optimization_mode_params['name']
		if name == 'reconstruction':
			return
		if name == 'knn_prediction':
			score, relation = report_knn_prediction(self.adata, self, self.optimization_mode_params,
													epoch, self.comet)
		elif name == 'modulation_prediction':
			score, relation = report_modulation_prediction(self.adata, self, self.optimization_mode_params,
														   epoch, self.comet)
		elif name == 'pseudo_metric':
			score, relation = report_pseudo_metric(self.adata, self, self.optimization_mode_params,
												   epoch, self.comet)
		else:
			raise ValueError('Unknown Optimization mode')
		if self.best_optimization_metric is None or relation(score, self.best_optimization_metric):
			self.best_optimization_metric = score
			self.save(os.path.join(save_path, f'best_model_by_metric.pt'))
		if self.comet is not None:
			self.comet.log_metric('max_metric', self.best_optimization_metric, epoch=epoch)

	def do_early_stopping(self, val_loss, early_stop, save_path, epoch):
		if self.best_loss is None or val_loss < self.best_loss:
			self.best_loss = val_loss
			self.save(os.path.join(save_path, 'best_model_by_reconstruction.pt'))
			self.no_improvements = 0
		else:
			self.no_improvements += 1
		if early_stop is not None and self.no_improvements > early_stop:
			print('Early stopped')
			return True
		if self.comet is not None:
			self.comet.log_metric('Epochs without Improvements', self.no_improvements, epoch=epoch)
		return False

	# <- prediction functions ->
	def get_latent(self, adata, metadata, return_mean=True):
		"""
		Get latent
		:param adata:
		:param metadata: list of str, list of metadata that is needed, not really useful at the moment
		:param return_mean: bool, calculate latent space without sampling
		:return: adata containing embedding vector in adata.X for each cell and the specified metadata in adata.obs
		"""
		data_embed = initialize_prediction_loader(adata, metadata, self.batch_size, beta_only=self.beta_only,
												  conditional=self.conditional)

		zs = []
		with torch.no_grad():
			self.model = self.model.to(self.device)
			self.model.eval()
			for rna, tcr, seq_len, _, labels, conditional in data_embed:
				rna = rna.to(self.device)
				tcr = tcr.to(self.device)

				if self.conditional is not None:
					conditional = conditional.to(self.device)
				else:
					conditional = None
				z, mu, _, _, _ = self.model(rna, tcr, seq_len, conditional)
				if return_mean:
					z = mu
				z = self.model.get_latent_from_z(z)
				z = sc.AnnData(z.detach().cpu().numpy())
				# z.obs[metadata] = np.array(metadata_batch).T
				zs.append(z)
		latent = sc.AnnData.concatenate(*zs)
		latent.obs.index = adata.obs.index
		latent.obs[metadata] = adata.obs[metadata]
		return latent

	def predict_rna_from_latent(self, adata_latent, metadata=None):
		data = initialize_latent_loader(adata_latent, self.batch_size, self.conditional)
		rnas = []
		with torch.no_grad():
			model = self.model.to(self.device)
			model.eval()
			for batch in data:
				if self.conditional is not None:
					batch = batch[0].to(self.device)
					conditional = batch[1].to(self.device)
				else:
					batch = batch[0].to(self.device)
					conditional = None
				batch_rna = model.predict_transcriptome(batch, conditional)
				batch_rna = sc.AnnData(batch_rna.detach().cpu().numpy())
				rnas.append(batch_rna)
		rnas = sc.AnnData.concatenate(*rnas)
		rnas.obs.index = adata_latent.obs.index
		if metadata is not None:
			rnas.obs[metadata] = adata_latent.obs[metadata]
		return rnas

	def predict_label(self, adata):
		data, _ = initialize_data_loader(adata, None, self.conditional, self.label_key,
										 None, self.batch_size, beta_only=self.beta_only)
		prediction_total = []
		with torch.no_grad():
			for rna, tcr, seq_len, metadata_batch, labels, conditional in data:
				rna = rna.to(self.device)
				tcr = tcr.to(self.device)

				if self.conditional is not None:
					conditional = conditional.to(self.device)
				else:
					conditional = None

				z = self.model(rna, tcr, seq_len, conditional)
				prediction = self.forward_supervised(z)
				prediction_total.append(prediction)
		prediction_total = torch.stack(prediction_total).cpu().detach().numpy()
		return prediction_total

	# <- semi-supervised model ->
	def forward_supervised(self, z):
		z_ = self.model.get_latent_from_z(z)
		prediction = self.supervised_model(z_)
		raise prediction

	# <- loss functions ->
	@abstractmethod
	def calculate_loss(self, rna_pred, rna, tcr_pred, tcr):
		raise NotImplementedError

	@abstractmethod
	def calculate_kld_loss(self, mu, logvar, epoch):
		"""
		Calculate the kld loss and z depending on the model type
		:param mu: mean of the VAE latent space
		:param logvar: log(variance) of the VAE latent space
		:param epoch: current epoch as integer
		:return:
		"""
		raise NotImplementedError('Implement this in the different model versions')

	def calculate_classification_loss(self, prediction, labels):
		loss = self.loss_function_class(prediction, labels)
		loss = self.loss_weights[3] * loss
		return loss

	def get_kl_annealing_factor(self, epoch):
		"""
		Calculate KLD annealing factor, i.e. KLD needs to get warmup
		:param epoch: current epoch
		:return:
		"""
		return min(1.0, epoch / self.kl_annealing_epochs)

	# <- logging helpers ->
	@property
	def history(self):
		return pd.DataFrame(self._val_history)

	@property
	def train_history(self):
		return pd.DataFrame(self._train_history)

	def save(self, filepath):
		""" Save model and optimizer state, and auxiliary data for continuing training """
		model_file = {'state_dict': self.model.state_dict(),
					  'train_history': self._train_history,
					  'val_history': self._val_history,
					  'aa_to_id': self.aa_to_id,

					  'params_architecture': self.params_architecture,
					  'balanced_sampling': self.balanced_sampling,
					  'metadata': self.metadata,
					  'conditional': self.conditional,
					  'optimization_mode_params': self.optimization_mode_params,
					  'label_key': self.label_key,
					  'model_type': self.model_type,
					  }
		torch.save(model_file, filepath)

	def load(self, filepath, map_location=torch.device('cuda')):
		""" Load model for evaluation / inference"""
		model_file = torch.load(os.path.join(filepath), map_location=map_location)
		self.model.load_state_dict(model_file['state_dict'], strict=False)
		self._train_history = model_file['train_history']
		self._val_history = model_file['val_history']
		self.aa_to_id = model_file['aa_to_id']



import torch
from torch import nn


class MSLE(nn.Module):
    def __init__(self):
        super().__init__()
        self.mse = nn.MSELoss()

    def forward(self, x_pred, x_true):
        log_pred = torch.log(x_pred+1)
        log_true = torch.log(x_true+1)
        loss = self.mse(log_pred, log_true)
        return loss
